"""Test implementation (interface realization) item connectors."""

from gaphor import UML
from gaphor.diagram.tests.fixtures import allow, connect, get_connected
from gaphor.UML.classes.interface import InterfaceItem
from gaphor.UML.classes.interfacerealization import InterfaceRealizationItem
from gaphor.UML.classes.klass import ClassItem


def test_non_interface_glue(create):
    """Test non-interface gluing with implementation."""

    impl = create(InterfaceRealizationItem)
    clazz = create(ClassItem, UML.Class)

    glued = allow(impl, impl.head, clazz)
    # connecting head to non-interface item is disallowed
    assert not glued


def test_interface_glue(create):
    """Test interface gluing with implementation."""
    iface = create(InterfaceItem, UML.Interface)
    impl = create(InterfaceRealizationItem)

    glued = allow(impl, impl.head, iface)
    assert glued


def test_classifier_glue(create):
    """Test classifier gluing with implementation."""
    impl = create(InterfaceRealizationItem)
    clazz = create(ClassItem, UML.Class)

    glued = allow(impl, impl.tail, clazz)
    assert glued


def test_connection(create):
    """Test connection of class and interface with implementation."""
    iface = create(InterfaceItem, UML.Interface)
    impl = create(InterfaceRealizationItem)
    clazz = create(ClassItem, UML.Class)

    connect(impl, impl.head, iface)
    connect(impl, impl.tail, clazz)

    # check the datamodel
    assert isinstance(impl.subject, UML.InterfaceRealization)
    ct = get_connected(impl, impl.head)
    assert ct is iface
    assert impl.subject is not None
    assert iface.subject == impl.subject.contract
    assert clazz.subject == impl.subject.implementingClassifier


def test_reconnection(create):
    """Test reconnection of class and interface with implementation."""
    iface = create(InterfaceItem, UML.Interface)
    c1 = create(ClassItem, UML.Class)
    c2 = create(ClassItem, UML.Class)
    impl = create(InterfaceRealizationItem)

    # connect: iface -> c1
    connect(impl, impl.head, iface)
    connect(impl, impl.tail, c1)

    s = impl.subject

    # reconnect: iface -> c2
    connect(impl, impl.tail, c2)

    assert s is not impl.subject
    assert iface.subject == impl.subject.contract
    assert c2.subject == impl.subject.implementingClassifier
    assert c1.subject != impl.subject.implementingClassifier



from io import StringIO

import pytest
from defusedxml import EntitiesForbidden

from gaphor.storage.parser import parse


def test_parsing_v2_1_model_with_grouped_item(test_models):
    with open(test_models / "node-component-v2.1.gaphor", encoding="utf-8") as f:
        elements = parse(f)

    diagram = next(e for e in elements.values() if e.type == "Diagram")
    component_item = next(e for e in elements.values() if e.type == "ComponentItem")
    node_item = next(e for e in elements.values() if e.type == "NodeItem")

    assert component_item.references["parent"] == node_item.id
    assert component_item.references["diagram"] == diagram.id
    assert "parent" not in node_item.references
    assert node_item.references["diagram"] == diagram.id


def test_parsing_of_open_file(test_models):
    with (test_models / "test-model.gaphor").open(encoding="utf-8") as model:
        elements = parse(model)

    assert elements


def test_parsing_of_xml_entities():
    model = StringIO(
        """<?xml version="1.0" encoding="utf-8"?>
        <!DOCTYPE gaphor [
        <!ENTITY ent "okay" >]>
        <gaphor xmlns="http://gaphor.sourceforge.net/model" version="3.0" gaphor-version="2.9.2">
         <Package id="0">
          <name>
           <val>&ent;</val>
          </name>
         </Package>
        </gaphor>"""
    )

    with pytest.raises(EntitiesForbidden):
        parse(model)


def test_parsing_of_xml_external_entities_should_fail():
    model = StringIO(
        """<?xml version="1.0" encoding="utf-8"?>
        <!DOCTYPE gaphor [
        <!ENTITY xxe SYSTEM "https://gaphor.org/latest.txt" >]>
        <gaphor xmlns="http://gaphor.sourceforge.net/model" version="3.0" gaphor-version="2.9.2">
         <Package id="0">
          <name>
           <val>&xxe;</val>
          </name>
         </Package>
        </gaphor>"""
    )

    with pytest.raises(EntitiesForbidden):
        parse(model)



import pytest

from gaphor.core.modeling import ElementFactory
from gaphor.core.modeling.elementdispatcher import ElementDispatcher
from gaphor.services.modelinglanguage import ModelingLanguageService
from gaphor.storage.parser import element
from gaphor.storage.tests.test_storage_upgrades import loader  # noqa: F401


@pytest.fixture
def modeling_language(event_manager):
    return ModelingLanguageService(event_manager)


@pytest.fixture
def element_factory(event_manager, modeling_language):
    element_factory = ElementFactory(
        event_manager, ElementDispatcher(event_manager, modeling_language)
    )
    yield element_factory
    element_factory.shutdown()


def test_c4_database_item(loader):  # noqa: F811
    parsed_item = element(id="2", type="C4ContainerDatabaseItem")
    item = loader(parsed_item)[0]

    assert item.__class__.__name__ == "C4DatabaseItem"



from gaphor import UML
from gaphor.core.modeling import Diagram, StyleSheet
from gaphor.UML.states.state import StateItem
from gaphor.UML.states.statemachine import StateMachineItem


def test_state(create):
    assert create(StateItem, UML.State)


def test_statemachine(create):
    assert create(StateMachineItem, UML.StateMachine)


def test_activities_persistence(create, element_factory, saver, loader):
    """Test state activities saving/loading."""
    # all activities
    s1 = create(StateItem, UML.State)
    s1.subject.name = "s1"
    s1.subject.entry = s1.model.create(UML.Activity)
    s1.subject.exit = s1.model.create(UML.Activity)
    s1.subject.doActivity = s1.model.create(UML.Activity)
    s1.subject.entry.name = "test 1 entry"
    s1.subject.exit.name = "test 1 exit"
    s1.subject.doActivity.name = "test 1 do"

    data = saver()
    loader(data)
    diagram = next(element_factory.select(Diagram))

    s1 = next(diagram.select(StateItem))
    assert "test 1 entry" == s1.subject.entry.name
    assert "test 1 exit" == s1.subject.exit.name
    assert "test 1 do" == s1.subject.doActivity.name


def test_state_machine_with_regions(create, diagram, element_factory):
    element_factory.create(StyleSheet)
    state_machine: StateMachineItem = create(StateMachineItem, UML.StateMachine)
    state_machine.subject.region = element_factory.create(UML.Region)
    state_machine.subject.region = element_factory.create(UML.Region)

    diagram.update()

    assert (
        state_machine.region_at_point((10, state_machine.height / 2))
        is state_machine.subject.region[0]
    )
    assert (
        state_machine.region_at_point((10, state_machine.height - 10))
        is state_machine.subject.region[1]
    )


def test_state_with_regions(create, diagram, element_factory):
    element_factory.create(StyleSheet)
    state: StateItem = create(StateItem, UML.State)
    state.subject.region = element_factory.create(UML.Region)
    state.subject.region = element_factory.create(UML.Region)

    diagram.update()

    assert state.region_at_point((10, state.height / 2)) is state.subject.region[0]
    assert state.region_at_point((10, state.height - 10)) is state.subject.region[1]



import itertools

import pytest

from gaphor.C4Model import diagramitems as c4_diagramitems
from gaphor.diagram.general import diagramitems as general_diagramitems
from gaphor.diagram.presentation import LinePresentation
from gaphor.diagram.support import get_diagram_item_metadata, get_model_element
from gaphor.RAAML import diagramitems as raaml_diagramitems
from gaphor.SysML import diagramitems as sysml_diagramitems
from gaphor.UML import diagramitems as uml_diagramitems
from gaphor.UML.uml import Element

diagramitems = [
    c
    for c in itertools.chain.from_iterable(
        diagramitems.__dict__.values()
        for diagramitems in (
            uml_diagramitems,
            sysml_diagramitems,
            raaml_diagramitems,
            c4_diagramitems,
            general_diagramitems,
        )
    )
    if isinstance(c, type)
]

blacklist = [
    # These lines have no subject of their own
    general_diagramitems.Line,
    uml_diagramitems.CommentLineItem,
    uml_diagramitems.ContainmentItem,
    # These lines have extra objects at the line ends
    uml_diagramitems.AssociationItem,
    uml_diagramitems.ConnectorItem,
    uml_diagramitems.ExtensionItem,
    uml_diagramitems.MessageItem,
]


@pytest.mark.parametrize(
    "diagram_item",
    [p for p in diagramitems if issubclass(p, LinePresentation) and p not in blacklist],
)
def test_line_presentations_have_metadata(diagram_item):
    metadata = get_diagram_item_metadata(diagram_item)
    element_class = get_model_element(diagram_item)

    assert metadata
    assert "head" in metadata
    assert "tail" in metadata
    assert issubclass(element_class, Element)
    assert metadata["head"] in element_class.__properties__
    assert metadata["tail"] in element_class.__properties__



import gc
from collections.abc import Container, Iterable

import pytest

from gaphor.core import event_handler
from gaphor.core.modeling.event import (
    ElementCreated,
    ElementDeleted,
    ModelChanged,
    ModelFlushed,
)
from gaphor.core.modeling.presentation import Presentation
from gaphor.core.modeling.stylesheet import StyleSheet
from gaphor.UML import LiteralString, Operation, Parameter


def test_element_factory_is_an_iterable(element_factory):
    assert isinstance(element_factory, Iterable)


def test_element_factory_is_a_container(element_factory):
    assert isinstance(element_factory, Container)


def test_create(element_factory):
    element_factory.create(Parameter)
    assert len(list(element_factory.values())) == 1


def test_create_is_idempotent(element_factory):
    param = element_factory.create(Parameter)
    new_param = element_factory.create_as(Parameter, param.id)
    assert param is new_param


def test_create_is_idempotent_but_validates_type(element_factory):
    param = element_factory.create(Parameter)
    with pytest.raises(TypeError):
        element_factory.create_as(Operation, param.id)


def test_should_not_create_presentation_elements(element_factory):
    with pytest.raises(TypeError):
        element_factory.create(Presentation)


def test_style_sheet(element_factory):
    style_sheet = element_factory.create(StyleSheet)

    assert element_factory.style_sheet is style_sheet


def test_flush(element_factory):
    p = element_factory.create(Parameter)
    assert len(list(element_factory.values())) == 1
    element_factory.flush()
    del p

    gc.collect()

    assert not list(element_factory.values()), list(element_factory.values())


def test_without_application(element_factory):
    element_factory.create(Parameter)
    assert element_factory.size() == 1, element_factory.size()

    element_factory.flush()
    assert element_factory.size() == 0, element_factory.size()

    p = element_factory.create(Parameter)
    assert element_factory.size() == 1, element_factory.size()

    p.unlink()
    assert element_factory.size() == 0, element_factory.size()


def test_unlink(element_factory):
    p = element_factory.create(Parameter)

    assert len(list(element_factory.values())) == 1

    p.unlink()

    assert not list(element_factory.values()), list(element_factory.values())

    p = element_factory.create(Parameter)
    defaultValue = element_factory.create(LiteralString)
    defaultValue.value = "l"
    defaultValue.owningParameter = p
    p.defaultValue = defaultValue

    assert len(list(element_factory.values())) == 2

    p.unlink()
    del p

    assert not list(element_factory.values()), list(element_factory.values())


# Event handlers are registered as persisting top level handlers, since no
# unsubscribe functionality is provided.
handled = False
events = []
last_event = None


@event_handler(ModelChanged)
def handler(event):
    global handled, events, last_event
    handled = True
    events.append(event)
    last_event = event


def clear_events():
    global handled, events, last_event
    handled = False
    events = []
    last_event = None


@pytest.fixture(autouse=True)
def subscribe_handlers(event_manager):
    event_manager.subscribe(handler)
    clear_events()
    yield None
    clear_events()


def test_create_event(element_factory):
    element_factory.create(Parameter)
    assert isinstance(last_event, ElementCreated)
    assert handled


def test_remove_event(element_factory):
    p = element_factory.create(Parameter)
    clear_events()
    p.unlink()
    assert isinstance(last_event, ElementDeleted)


def test_flush_event(element_factory):
    element_factory.create(Parameter)
    del events[:]
    element_factory.flush()
    assert len(events) == 1, events
    assert isinstance(last_event, ModelFlushed)


def test_no_create_events_when_blocked(element_factory):
    with element_factory.block_events():
        element_factory.create(Parameter)
    assert events == [], events


class TriggerUnlink:
    def __init__(self, element):
        self.element = element


class CheckModel:
    def __init__(self, element):
        self.element = element


def test_indirect_delete_of_element(event_manager, element_factory):
    @event_handler(CheckModel)
    def on_check_model(event):
        assert event.element.model
        assert event.element in element_factory

    @event_handler(TriggerUnlink)
    def on_trigger_unlink(event):
        event_manager.handle(CheckModel(event.element))
        event.element.unlink()

    event_manager.subscribe(on_check_model)
    event_manager.subscribe(on_trigger_unlink)

    operation = element_factory.create(Operation)
    event_manager.handle(TriggerUnlink(operation))

    with pytest.raises(TypeError):
        assert operation.model
    assert operation not in element_factory



from gaphor import UML
from gaphor.diagram.tests.fixtures import find
from gaphor.UML.actions.actionspropertypages import (
    DecisionNodePropertyPage,
    FlowPropertyPageAbstract,
    ForkNodePropertyPage,
    JoinNodePropertyPage,
    ObjectNodePropertyPage,
    ShowObjectNodePropertyPage,
)


def test_object_node_property_page_show_ordering(
    diagram, element_factory, event_manager
):
    item = diagram.create(
        UML.actions.ObjectNodeItem, subject=element_factory.create(UML.ObjectNode)
    )
    property_page = ShowObjectNodePropertyPage(item, event_manager)

    widget = property_page.construct()
    show_ordering = find(widget, "show-ordering")
    show_ordering.set_active(True)

    assert item.show_ordering


def test_object_node_property_page_upper_bound(diagram, element_factory, event_manager):
    subject = element_factory.create(UML.ObjectNode)
    property_page = ObjectNodePropertyPage(subject, event_manager)

    widget = property_page.construct()
    upper_bound = find(widget, "upper-bound")
    upper_bound.set_text("test")

    assert subject.upperBound.value == "test"


def test_object_node_property_page_ordering(diagram, element_factory, event_manager):
    subject = element_factory.create(UML.ObjectNode)

    assert subject.ordering == "FIFO"
    property_page = ObjectNodePropertyPage(subject, event_manager)

    widget = property_page.construct()
    ordering = find(widget, "ordering")
    ordering.set_selected(0)

    assert subject.ordering == "unordered"


def test_decision_node_property_page_show_type(diagram, element_factory, event_manager):
    item = diagram.create(
        UML.actions.DecisionNodeItem, subject=element_factory.create(UML.DecisionNode)
    )
    property_page = DecisionNodePropertyPage(item, event_manager)

    widget = property_page.construct()
    show_type = find(widget, "show-type")
    show_type.set_active(True)

    assert item.show_underlying_type


def test_fork_node_property_page(diagram, element_factory, event_manager):
    item = diagram.create(
        UML.actions.ForkNodeItem, subject=element_factory.create(UML.ForkNode)
    )
    orig_matrix = item.matrix.tuple()
    property_page = ForkNodePropertyPage(item, event_manager)

    widget = property_page.construct()
    horizontal = find(widget, "horizontal")
    horizontal.set_active(True)

    assert item.matrix.tuple() != orig_matrix


def test_join_node_property_page(element_factory, event_manager):
    subject = element_factory.create(UML.JoinNode)
    property_page = JoinNodePropertyPage(subject, event_manager)

    widget = property_page.construct()
    join_spec = find(widget, "join-spec")
    join_spec.set_text("test")

    assert subject.joinSpec.value == "test"


def test_flow_property_page(element_factory, event_manager):
    subject = element_factory.create(UML.ObjectFlow)
    property_page = FlowPropertyPageAbstract(subject, event_manager)

    widget = property_page.construct()
    guard = find(widget, "guard")
    guard.set_text("test")

    assert subject.guard.value == "test"



from gi.repository import Gdk, GLib

from gaphor.plugins.console.console import GTKInterpreterConsole, Help, main


class KeyEvent:
    def __init__(self, keyval, state=0):
        self.keyval = keyval
        self.state = state


class DummyWriter:
    def __init__(self):
        self.text = ""

    def write(self, s):
        self.text = self.text + s

    def __enter__(self):
        return self

    def __exit__(self, _type, _value, _traceback):
        pass


def iteration():
    ctx = GLib.main_context_default()
    while ctx.pending():
        ctx.iteration(False)


def test_console_opening():
    main(main_loop=False)
    iteration()


def test_help():
    writer = DummyWriter()
    help = Help(writer)

    help()

    assert "Usage: help(object)" in writer.text


def console_text(console):
    buffer = console.text.get_buffer()
    return buffer.get_text(buffer.get_start_iter(), buffer.get_end_iter(), False)


def test_history():
    console = GTKInterpreterConsole(locals={})
    console.history.append("my_history()")

    console.key_pressed(console, Gdk.keyval_from_name("Up"), -1, 0)
    text = console_text(console)

    assert ">>> my_history()" in text


def test_deep_history():
    console = GTKInterpreterConsole(locals={})
    console.history.append("deepest()")
    console.history.append("deeper()")
    console.history.append("deep()")

    console.key_pressed(console, Gdk.keyval_from_name("Up"), -1, 0)
    console.key_pressed(console, Gdk.keyval_from_name("Up"), -1, 0)
    console.key_pressed(console, Gdk.keyval_from_name("Up"), -1, 0)
    console.key_pressed(console, Gdk.keyval_from_name("Down"), -1, 0)
    text = console_text(console)

    assert ">>> deeper()" in text


def test_run_line():
    console = GTKInterpreterConsole(locals={})

    console.buffer.append("help")
    console.key_pressed(console, Gdk.keyval_from_name("Return"), -1, 0)

    text = console_text(console)

    assert "Usage: help(object)" in text



"""Tests for SVG overlay injection."""

import xml.etree.ElementTree as ET

import pytest

from gaphor import UML
from gaphor.diagram.export import save_svg
from gaphor.diagram.general.simpleitem import Box
from gaphor.diagram.tests.fixtures import connect
from gaphor.plugins.htmlreport.svg_overlay import inject_overlays, item_svg_bounds
from gaphor.UML.classes import ClassItem
from gaphor.UML.classes.association import AssociationItem


@pytest.fixture
def diagram_with_classes(diagram, element_factory, create):
    c1 = create(ClassItem, UML.Class)
    c1.subject.name = "ClassA"
    c2 = create(ClassItem, UML.Class)
    c2.subject.name = "ClassB"

    # Position them apart
    c1.matrix.translate(50, 50)
    c2.matrix.translate(250, 50)
    diagram.update(diagram.ownedPresentation)

    return diagram, c1, c2


def test_item_svg_bounds_no_subject(diagram):
    box = diagram.create(Box)
    diagram.update(diagram.ownedPresentation)
    bounds = item_svg_bounds(box, 0, 0)
    assert bounds is None


def test_inject_overlays_adds_links(diagram_with_classes, tmp_path):
    diagram, c1, c2 = diagram_with_classes
    svg_path = tmp_path / "test.svg"
    save_svg(str(svg_path), diagram)
    inject_overlays(str(svg_path), diagram)

    content = svg_path.read_text(encoding="utf-8")
    assert f"#element/{c1.subject.id}" in content
    assert f"#element/{c2.subject.id}" in content

    # Parse and verify structure
    tree = ET.parse(str(svg_path))
    root = tree.getroot()
    ns = {"svg": "http://www.w3.org/2000/svg"}
    links = root.findall(".//svg:a", ns)
    assert len(links) >= 2

    # Check that rects are present inside links
    rects = root.findall(".//svg:a/svg:rect", ns)
    assert len(rects) >= 2


def test_inject_overlays_with_line(diagram, element_factory, create, tmp_path):
    c1 = create(ClassItem, UML.Class)
    c1.subject.name = "Src"
    c1.matrix.translate(50, 50)

    c2 = create(ClassItem, UML.Class)
    c2.subject.name = "Dst"
    c2.matrix.translate(300, 50)

    assoc = create(AssociationItem)
    connect(assoc, assoc.head, c1)
    connect(assoc, assoc.tail, c2)

    diagram.update(diagram.ownedPresentation)

    svg_path = tmp_path / "test_line.svg"
    save_svg(str(svg_path), diagram)
    inject_overlays(str(svg_path), diagram)

    # Lines with subjects should have polyline overlays
    ns = {"svg": "http://www.w3.org/2000/svg"}
    tree = ET.parse(str(svg_path))
    root = tree.getroot()
    links = root.findall(".//svg:a", ns)
    # At least 2 class rects + potentially the association line
    assert len(links) >= 2

    polylines = root.findall(".//svg:a/svg:polyline", ns)
    assert len(polylines) >= 1



import logging

from gaphor.abc import Service

log = logging.getLogger(__name__)


class TestService(Service):
    def __init__(self):
        log.info("Initializing test service")

    def shutdown(self):
        log.info("Shutting down test service")



import pytest

from gaphor.diagram.copypaste import copy, paste


def test_copy_unknown_type_raises_exception():
    with pytest.raises(ValueError):
        copy(None)


def test_paste_unknown_type_raises_exception(diagram, element_factory):
    with pytest.raises(ValueError):
        paste(None, diagram, element_factory.lookup)



import pytest
from gi.repository import Gtk

from gaphor.services.undomanager import UndoManager
from gaphor.ui.undoactions import UndoActions


class MockMainWindow:
    def __init__(self):
        self.window = Gtk.Window()


@pytest.fixture
def main_window():
    return MockMainWindow()


@pytest.fixture
def undo_manager(event_manager, element_factory):
    undo_manager = UndoManager(event_manager, element_factory)
    yield undo_manager
    undo_manager.shutdown()


@pytest.fixture
def undo_actions(undo_manager, main_window):
    return UndoActions(undo_manager, main_window)


@pytest.mark.parametrize(
    "focus,action", [[True, "undo"], [True, "redo"], [False, "undo"], [False, "redo"]]
)
def test_undo_action(undo_actions, main_window, monkeypatch, focus, action):
    activated_action = None
    undo_manager_called = False

    gtk_text_activate_action = Gtk.Text.activate_action

    def mock_activate_action(self, name):
        nonlocal activated_action
        activated_action = name
        return gtk_text_activate_action(self, name)

    def mock_undo_transaction(self):
        nonlocal undo_manager_called
        undo_manager_called = True

    monkeypatch.setattr(Gtk.Text, "activate_action", mock_activate_action)
    monkeypatch.setattr(UndoManager, f"{action}_transaction", mock_undo_transaction)

    win = main_window.window
    entry = Gtk.Entry.new()
    win.set_child(entry)
    win.present()
    win.set_focus(entry if focus else None)

    getattr(undo_actions, f"{action}_transaction")()

    if focus:
        assert activated_action == f"text.{action}"
        assert not undo_manager_called
    else:
        assert not activated_action
        assert undo_manager_called



import locale

from gaphor import UML
from gaphor.UML.actions.objectnode import ObjectNodeItem


def test_object_node(diagram, element_factory):
    node = element_factory.create(UML.ObjectNode)
    item = diagram.create(ObjectNodeItem, subject=node)

    assert item.subject is node


def test_name(create):
    node = create(ObjectNodeItem, UML.ObjectNode)
    name = node.shape.icon.child.children[1]

    node.subject.name = "Blah"

    assert "Blah" == name.child.text()


def test_ordering(create):
    """Test updating of ObjectNodeItem.ordering."""
    node = create(ObjectNodeItem, UML.ObjectNode)
    ordering = node.shape.children[1]

    node.subject.ordering = "unordered"
    node.show_ordering = True

    language = locale.getlocale()[0]
    if language == "en_US":
        assert "{ ordering = unordered }" == ordering.child.text()



from gaphas.handlemove import HandleMove

from gaphor.core import event_handler
from gaphor.diagram.connectors import ItemTemporaryDisconnected
from gaphor.diagram.presentation import LinePresentation
from gaphor.diagram.tests.fixtures import connect
from gaphor.UML import Comment
from gaphor.UML.general import CommentItem, CommentLineItem


def test_handle_move_has_guides(create, view):
    line = create(LinePresentation)

    handle_move = HandleMove(line, line.handles()[0], view)

    handle_move.start_move((0, 0))
    handle_move.move((0, 0))

    assert view.guides


def test_handle_move_has_gray_out(create, view):
    line = create(CommentLineItem)
    line2 = create(CommentLineItem)

    handle_move = HandleMove(line, line.handles()[0], view)

    handle_move.start_move((0, 0))

    assert line2 in view.selection.grayed_out_items


def test_handle_move_sends_disconnect_event(create, view, event_manager):
    comment = create(CommentItem, element_class=Comment)
    line = create(CommentLineItem)
    connect(line, line.handles()[0], comment)

    events = []

    @event_handler(ItemTemporaryDisconnected)
    def handler(e):
        events.append(e)

    event_manager.subscribe(handler)

    handle_move = HandleMove(line, line.handles()[0], view)
    handle_move.start_move((0, 0))

    assert events



import pytest

from gaphor.core.styling import merge_styles
from gaphor.core.styling.declarations import Var


def test_merge_opacity():
    style = merge_styles({"color": (0, 0, 1, 1), "opacity": 0.7})

    assert style["color"] == (0, 0, 1, 0.7)


def test_merge_opacity_with_transparent():
    style = merge_styles({"color": (0, 0, 0, 0), "opacity": 0.7})

    assert style["color"] == (0, 0, 0, 0)


def test_merge_opacity_with_transparency():
    style = merge_styles({"color": (0, 0, 0, 0.8), "opacity": 0.5})

    assert style["color"] == (0, 0, 0, 0.4)


@pytest.mark.parametrize(
    "font_size,factor",
    [
        ["x-small", 3 / 4],
        ["small", 8 / 9],
        ["medium", 1],
        ["large", 6 / 5],
        ["x-large", 3 / 2],
    ],
)
def test_font_size(font_size, factor):
    style = merge_styles({"font-size": 10}, {"font-size": font_size})

    assert style["font-size"] == 10 * factor


def test_final_font_size_is_a_number():
    style = merge_styles({"font-size": 10}, {"font-size": "x-small"}, {"font-size": 24})

    assert style["font-size"] == 24


def test_font_size_override_with_relative_size():
    style = merge_styles({"font-size": 10}, {"font-size": 24}, {"font-size": "x-small"})

    assert style["font-size"] == 24 * 3 / 4


def test_color_override_with_variables():
    style = merge_styles(
        {"--my-color": "white", "--my-opacity": 0.5},
        {"color": Var("--my-color"), "opacity": Var("--my-opacity")},
    )

    assert style["color"] == (1.0, 1.0, 1.0, 0.5)



from gaphor import UML
from gaphor.core.modeling import StyleSheet
from gaphor.core.modeling.diagram import StyledDiagram
from gaphor.UML.classes import ClassItem


def test_sub_expression(diagram, create, element_factory):
    style_sheet = element_factory.create(StyleSheet)
    style_sheet.styleSheet = ":root:has(class name) { color: blue; }"
    create(ClassItem, UML.Class)

    compiled_style_sheet = style_sheet.compile_style_sheet()
    props = compiled_style_sheet.compute_style(StyledDiagram(diagram))

    assert props["color"] == (0, 0, 1, 1)



from gaphor import UML
from gaphor.SysML.blocks.block import BlockItem
from gaphor.SysML.sysml import Block
from gaphor.UML.recipes import apply_stereotype


def test_block_item_show_own_stereotype(create):
    block_item = create(BlockItem, Block)

    assert "block" in block_stereotype_text(block_item)


def test_block_item_with_stereotype(element_factory, create):
    stereotype = element_factory.create(UML.Stereotype)
    stereotype.name = "MyStereotype"
    block_item = create(BlockItem, Block)
    apply_stereotype(block_item.subject, stereotype)

    assert "block" not in block_stereotype_text(block_item)
    assert "myStereotype" in block_stereotype_text(block_item)


def block_stereotype_text(block_item):
    return block_item.shape.children[0].child.children[0].child.text()



from gaphor import UML
from gaphor.UML.states.transition import TransitionItem


def test_transition_guard(create, element_factory):
    """Test events of transition.guard."""
    item = create(TransitionItem, UML.Transition)
    guard = item.shape_middle
    assert guard.child.text() == ""

    c = element_factory.create(UML.Constraint)
    specification = element_factory.create(UML.LiteralString)
    specification.owningConstraint = c
    specification.value = "blah"
    c.specification = specification
    assert guard.child.text() == ""

    item.subject.guard = c
    assert item.subject.guard is c
    assert guard.child.text() == "[blah]", guard.text()

    del c.specification
    assert guard.child.text() == "", guard.text()

    specification.value = "foo"
    c.specification = specification
    assert guard.child.text() == "[foo]", guard.text()



"""Test undo, in particular for item specific properties:

* Line: orthogonal, horizontal
* Matrix updates
* Variables
* Solver: add_constraint, remove_constraint
"""

import pytest
from gaphas.connector import Handle
from gaphas.segment import Segment

from gaphor.core import Transaction
from gaphor.diagram.presentation import ElementPresentation, LinePresentation
from gaphor.diagram.tests.fixtures import connect
from gaphor.UML import Class
from gaphor.UML.classes import ClassItem, GeneralizationItem


def test_line_create(diagram, undo_manager, event_manager):
    with Transaction(event_manager):
        diagram.create(LinePresentation)

    assert diagram.ownedPresentation

    undo_manager.undo_transaction()

    assert not diagram.ownedPresentation

    undo_manager.redo_transaction()

    assert diagram.ownedPresentation


def test_line_delete(diagram, undo_manager, event_manager):
    with Transaction(event_manager):
        line = diagram.create(LinePresentation)
        line.insert_handle(1, Handle((20, 20)))
        line.matrix.translate(10, 10)

    with Transaction(event_manager):
        line.unlink()

    undo_manager.undo_transaction()

    line = diagram.ownedPresentation[0]
    assert len(line.handles()) == 3
    assert line.handles()[1].pos.tuple() == (20, 20)
    assert line.matrix.tuple() == (1, 0, 0, 1, 10, 10)


def test_line_orthogonal_property(diagram, undo_manager, event_manager):
    with Transaction(event_manager):
        line = diagram.create(LinePresentation)
        line.insert_handle(0, Handle())

    with Transaction(event_manager):
        line.orthogonal = True

    assert line.orthogonal

    undo_manager.undo_transaction()

    assert not line.orthogonal

    undo_manager.redo_transaction()

    assert line.orthogonal


def test_line_horizontal_property(diagram, undo_manager, event_manager):
    with Transaction(event_manager):
        line = diagram.create(LinePresentation)
        line.insert_handle(0, Handle())

    with Transaction(event_manager):
        line.horizontal = True

    assert line.horizontal

    undo_manager.undo_transaction()

    assert not line.horizontal

    undo_manager.redo_transaction()

    assert line.horizontal


@pytest.mark.parametrize(
    "action",
    [
        lambda line: line.matrix.translate(10, 10),
        lambda line: line.matrix.rotate(1),
        lambda line: line.matrix.scale(2.5, 3.5),
        lambda line: line.matrix.invert(),
    ],
)
def test_matrix_operation(action, diagram, undo_manager, event_manager):
    with Transaction(event_manager):
        line = diagram.create(LinePresentation)
        line.matrix.translate(10, 0)

    original = tuple(line.matrix)

    with Transaction(event_manager):
        action(line)

    assert tuple(line.matrix) != original

    undo_manager.undo_transaction()

    assert tuple(line.matrix) == original

    undo_manager.redo_transaction()

    assert tuple(line.matrix) != original


@pytest.mark.parametrize("index", range(4))
def test_element_handle_position(diagram, undo_manager, event_manager, index):
    with Transaction(event_manager):
        element = diagram.create(ElementPresentation)

    handle = element.handles()[index]
    old_pos = handle.pos.tuple()
    new_pos = (30, 40)

    with Transaction(event_manager):
        handle.pos = new_pos

    assert tuple(handle.pos) == new_pos

    undo_manager.undo_transaction()

    assert handle.pos.tuple() == old_pos

    undo_manager.redo_transaction()

    assert tuple(handle.pos) == new_pos


@pytest.mark.parametrize("index", range(2))
def test_line_handle_position(diagram, undo_manager, event_manager, index):
    with Transaction(event_manager):
        line = diagram.create(LinePresentation)

    handle = line.handles()[index]
    old_pos = handle.pos.tuple()
    new_pos = (30, 40)

    with Transaction(event_manager):
        handle.pos = new_pos

    assert tuple(handle.pos) == new_pos

    undo_manager.undo_transaction()

    assert handle.pos.tuple() == old_pos

    undo_manager.redo_transaction()

    assert tuple(handle.pos) == new_pos


def test_line_handle_on_inserted_handle(diagram, undo_manager, event_manager):
    with Transaction(event_manager):
        line = diagram.create(LinePresentation)

    handle = Handle()
    line.insert_handle(1, handle)

    old_pos = handle.pos.tuple()
    new_pos = (30, 40)

    with Transaction(event_manager):
        handle.pos = new_pos

    assert tuple(handle.pos) == new_pos

    undo_manager.undo_transaction()

    assert handle.pos.tuple() == old_pos

    undo_manager.redo_transaction()

    assert tuple(handle.pos) == new_pos


def test_line_handle_no_events_for_removed_handle(diagram, undo_manager, event_manager):
    with Transaction(event_manager):
        line = diagram.create(LinePresentation)

    handle = Handle()
    line.insert_handle(1, handle)
    line.remove_handle(handle)

    new_pos = (30, 40)

    with Transaction(event_manager):
        handle.pos = new_pos

    assert tuple(handle.pos) == new_pos

    undo_manager.undo_transaction()

    assert handle.pos.tuple() == new_pos


def test_line_loading_of_points(diagram, undo_manager, event_manager, element_factory):
    with Transaction(event_manager):
        line = diagram.create(LinePresentation)
        line.load("points", "[(0, 0), (5, 5), (10, 10)]")
        assert len(line.handles()) == 3

    handle = line.handles()[1]
    old_pos = handle.pos.tuple()
    new_pos = (30, 40)

    with Transaction(event_manager):
        handle.pos = new_pos

    assert tuple(handle.pos) == new_pos

    undo_manager.undo_transaction()

    assert handle.pos.tuple() == old_pos

    undo_manager.redo_transaction()

    assert tuple(handle.pos) == new_pos


def test_line_connections(diagram, undo_manager, element_factory, event_manager):
    with Transaction(event_manager):
        class_item = diagram.create(ClassItem, subject=element_factory.create(Class))
        gen_item = diagram.create(GeneralizationItem)

    handle = gen_item.handles()[0]

    with Transaction(event_manager):
        connect(gen_item, handle, class_item)

    connections = diagram.connections

    assert connections.get_connection(handle)

    undo_manager.undo_transaction()

    assert not connections.get_connection(handle)

    undo_manager.redo_transaction()

    assert connections.get_connection(handle)


def test_line_split_segment(diagram, undo_manager, event_manager):
    with Transaction(event_manager):
        line = diagram.create(LinePresentation)

    head_handle = line.head
    tail_handle = line.tail

    with Transaction(event_manager):
        segment = Segment(line, diagram)
        segment.split((5, 5))

    assert len(line.handles()) == 3

    undo_manager.undo_transaction()

    assert len(line.handles()) == 2
    assert line.head is head_handle
    assert line.tail is tail_handle

    undo_manager.redo_transaction()

    assert len(line.handles()) == 3


def test_line_merge_segment(diagram, undo_manager, event_manager):
    with Transaction(event_manager):
        line = diagram.create(LinePresentation)
        segment = Segment(line, diagram)
        segment.split((5, 5))

    head_handle = line.head
    tail_handle = line.tail

    with Transaction(event_manager):
        segment = Segment(line, diagram)
        segment.merge_segment(0)

    assert len(line.handles()) == 2
    assert line.head is head_handle
    assert line.tail is tail_handle

    undo_manager.undo_transaction()

    assert len(line.handles()) == 3
    assert line.head is head_handle
    assert line.tail is tail_handle

    undo_manager.redo_transaction()

    assert len(line.handles()) == 2


def test_undo_nested_element(diagram, undo_manager, event_manager):
    with Transaction(event_manager):
        parent = diagram.create(ElementPresentation)
        child = diagram.create(ElementPresentation)
        child.change_parent(parent)
        child.matrix.translate(100, 100)
        parent.matrix.translate(100, 100)

    with Transaction(event_manager):
        parent.unlink()

    undo_manager.undo_transaction()

    new_parent, new_child = diagram.ownedPresentation

    assert new_child.parent is new_parent
    assert tuple(new_parent.matrix_i2c) == (1.0, 0, 0, 1, 100, 100)
    assert tuple(new_child.matrix_i2c) == (1.0, 0, 0, 1, 200, 200)



import sphinx.application
import sphinx.util.docutils

from gaphor.extensions.sphinx import setup as sphinx_setup


def test_setup(tmp_path):
    (tmp_path / "conf.py").write_text("", encoding="utf-8")
    gen = sphinx.application.Sphinx(
        srcdir="docs",
        confdir=str(tmp_path),
        outdir=str(tmp_path),
        doctreedir=str(tmp_path),
        buildername="html",
    )

    result = sphinx_setup(gen)

    assert result["version"] == "0.1"
    assert result["parallel_read_safe"]
    assert result["parallel_write_safe"]

    assert sphinx.util.docutils.is_directive_registered("diagram")
    assert gen.config.gaphor_models == {}


def test_setup_load_models(tmp_path):
    (tmp_path / "conf.py").write_text(
        "gaphor_models = {'style-sheets': 'style-sheet-examples.gaphor'}",
        encoding="utf-8",
    )
    gen = sphinx.application.Sphinx(
        srcdir="docs",
        confdir=tmp_path,
        outdir=tmp_path,
        doctreedir=tmp_path,
        buildername="html",
    )

    result = sphinx_setup(gen)

    assert result["version"] == "0.1"
    assert result["parallel_read_safe"]
    assert result["parallel_write_safe"]

    assert sphinx.util.docutils.is_directive_registered("diagram")



from gaphor import UML
from gaphor.diagram.tests.fixtures import find
from gaphor.UML.classes.associationpropertypages import (
    AssociationDirectionPropertyPage,
    AssociationPropertyPage,
)


def test_association_property_page(element_factory, event_manager):
    end1 = element_factory.create(UML.Class)
    end2 = element_factory.create(UML.Class)
    subject = UML.recipes.create_association(end1, end2)

    property_page = AssociationPropertyPage(subject, event_manager)

    widget = property_page.construct()
    head_name = find(widget, "head-name")
    head_name.set_text("head")

    assert subject.memberEnd[0].name == "head"


def test_association_property_page_invert_direction(
    diagram, element_factory, event_manager
):
    end1 = element_factory.create(UML.Class)
    end2 = element_factory.create(UML.Class)
    item = diagram.create(
        UML.classes.AssociationItem, subject=UML.recipes.create_association(end1, end2)
    )
    item.head_subject = item.subject.memberEnd[0]
    item.tail_subject = item.subject.memberEnd[1]
    property_page = AssociationDirectionPropertyPage(item, event_manager)

    property_page.on_invert_direction_change(None)

    assert item.tail_subject is item.subject.memberEnd[0]


def metaclass_and_stereotype(element_factory):
    metaclass = element_factory.create(UML.Class)
    metaclass.name = "Element"
    stereotype = element_factory.create(UML.Stereotype)
    stereotype.name = "Stereotype"
    UML.recipes.create_extension(metaclass, stereotype)
    return metaclass, stereotype


def test_association_property_with_stereotype(element_factory, event_manager):
    end1 = element_factory.create(UML.Class)
    end2 = element_factory.create(UML.Class)
    subject = UML.recipes.create_association(end1, end2)
    metaclass_and_stereotype(element_factory)

    property_page = AssociationPropertyPage(subject, event_manager)

    widget = property_page.construct()

    assert widget



from gaphor.i18n import translated_ui_string


def test_ui_builder_builds_from_resources():
    ui_xml = translated_ui_string("gaphor.diagram", "propertypages.ui")

    assert "translatable=" not in ui_xml



"""Diagram could not be loaded due to JuggleError (presumed cyclic resolving of
diagram items)."""

from gaphor import UML
from gaphor.storage.load import load
from gaphor.UML.classes.association import (
    draw_head_navigable,
    draw_tail_composite,
    draw_tail_shared,
)


def test_association_ends_are_set(element_factory, modeling_language, test_models):
    with (test_models / "association-ends.gaphor").open(encoding="utf-8") as file_obj:
        load(file_obj, element_factory, modeling_language)
    composite = next(
        element_factory.select(
            lambda e: isinstance(e, UML.Association) and e.name == "composite"
        )
    )
    shared = next(
        element_factory.select(
            lambda e: isinstance(e, UML.Association) and e.name == "shared"
        )
    )

    assert composite.presentation[0].draw_head is draw_head_navigable
    assert composite.presentation[0].draw_tail is draw_tail_composite
    assert shared.presentation[0].draw_head is draw_head_navigable
    assert shared.presentation[0].draw_tail is draw_tail_shared



import contextlib

import pytest

import gaphor.C4Model.diagramitems
import gaphor.diagram.general.diagramitems
import gaphor.RAAML.diagramitems
import gaphor.SysML.diagramitems
import gaphor.UML.diagramitems
from gaphor.core.modeling import Base
from gaphor.core.modeling.properties import umlproperty


def presentations(module):
    for name in dir(module):
        element_type = getattr(module, name)
        with contextlib.suppress(TypeError):
            if issubclass(element_type, Base):
                yield element_type


def flatten(list):
    return [item for sublist in list for item in sublist]


def test_extract_presentations():
    assert list(presentations(gaphor.UML.diagramitems))


@pytest.mark.parametrize(
    "element_type",
    set(
        flatten(
            [
                presentations(gaphor.diagram.general.diagramitems),
                presentations(gaphor.UML.diagramitems),
                presentations(gaphor.SysML.diagramitems),
                presentations(gaphor.C4Model.diagramitems),
                presentations(gaphor.RAAML.diagramitems),
            ]
        )
    ),
)
def test_property_name_matches_assigned_name(element_type):
    for name in dir(element_type):
        prop = getattr(element_type, name)
        if isinstance(prop, umlproperty):
            assert name == prop.name, (
                f"No matching property name for {element_type}.{name}"
            )



"""Test issues where associations are copied and pasted, deleted, etc.

Scenario's:
* Class and association are pasted in a new diagram
* Class and association are pasted in a new diagram and original association is deleted
* Class and association are pasted in a new diagram and new association is deleted
* Association is pasted in a new diagram and reconnected to the class (same subject as original)
* Association is pasted and directly deleted
* Class and association are pasted in a new diagram and one end is connected to a different class

What's the behavior we're looking for?

* If an association has >1 presentations, it can only connect to those same types ()
* If an association has ==1 presentation, it can reconnect to another type, association is updated
"""

import pytest

from gaphor import UML
from gaphor.application import Session
from gaphor.core import Transaction
from gaphor.core.modeling import Diagram
from gaphor.diagram.copypaste import copy_full, paste_link
from gaphor.diagram.tests.fixtures import connect
from gaphor.UML import diagramitems
from gaphor.UML.recipes import set_navigability


@pytest.fixture
def session():
    session = Session()
    yield session
    session.shutdown()


@pytest.fixture
def event_manager(session):
    return session.get_service("event_manager")


@pytest.fixture
def element_factory(session):
    return session.get_service("element_factory")


@pytest.fixture
def diagram(event_manager, element_factory):
    with Transaction(event_manager):
        return element_factory.create(UML.Diagram)


@pytest.fixture
def class_and_association_with_copy(diagram, event_manager, element_factory):
    with Transaction(event_manager):
        c = diagram.create(
            diagramitems.ClassItem, subject=element_factory.create(UML.Class)
        )

        a = diagram.create(diagramitems.AssociationItem)
        connect(a, a.handles()[0], c)
        connect(a, a.handles()[1], c)

        set_navigability(a.subject, a.subject.memberEnd[0], True)

        copy_buffer = copy_full({a, c})
        new_diagram = element_factory.create(Diagram)
        pasted_items = paste_link(copy_buffer, new_diagram)

    aa = pasted_items.pop()
    if not isinstance(aa, diagramitems.AssociationItem):
        aa = pasted_items.pop()

    return c, a, aa


def test_delete_copied_associations(class_and_association_with_copy, event_manager):
    c, a, aa = class_and_association_with_copy

    assert a.subject.memberEnd[0].type
    assert a.subject.memberEnd[1].type
    assert a.subject.memberEnd[0].type is c.subject
    assert a.subject.memberEnd[1].type is c.subject
    assert a.subject.memberEnd[0] is a.head_subject
    assert a.subject.memberEnd[1] is a.tail_subject
    assert a.subject.memberEnd[0] in a.subject.memberEnd[1].type.ownedAttribute

    # Delete the copy and all is fine

    with Transaction(event_manager):
        aa.unlink()

    assert a.subject.memberEnd[0].type
    assert a.subject.memberEnd[1].type
    assert a.subject.memberEnd[0].type is c.subject
    assert a.subject.memberEnd[1].type is c.subject
    assert a.subject.memberEnd[0] is a.head_subject
    assert a.subject.memberEnd[1] is a.tail_subject
    assert a.subject.memberEnd[0] in a.subject.memberEnd[1].type.ownedAttribute


def test_delete_original_association(class_and_association_with_copy, event_manager):
    c, a, aa = class_and_association_with_copy

    assert aa.subject.memberEnd[0].type
    assert aa.subject.memberEnd[1].type
    assert aa.subject.memberEnd[0].type is c.subject
    assert aa.subject.memberEnd[1].type is c.subject
    assert aa.subject.memberEnd[0] is aa.head_subject
    assert aa.subject.memberEnd[1] is aa.tail_subject
    assert aa.subject.memberEnd[0] in aa.subject.memberEnd[1].type.ownedAttribute

    # Now, when the original is deleted, the model is changed and made invalid

    with Transaction(event_manager):
        a.unlink()

    assert aa.subject.memberEnd[0].type
    assert aa.subject.memberEnd[1].type
    assert aa.subject.memberEnd[0].type is c.subject
    assert aa.subject.memberEnd[1].type is c.subject
    assert aa.subject.memberEnd[0] is aa.head_subject
    assert aa.subject.memberEnd[1] is aa.tail_subject
    assert aa.subject.memberEnd[0] in aa.subject.memberEnd[1].type.ownedAttribute



import pytest

from gaphor import UML
from gaphor.diagram.instanteditors import named_item_editor


@pytest.mark.asyncio
async def test_named_item_editor_with_element(
    diagram, element_factory, view, event_manager
):
    item = diagram.create(
        UML.classes.ClassItem, subject=element_factory.create(UML.Class)
    )
    view.selection.hovered_item = item
    await view.update()
    result = named_item_editor(item, view, event_manager)

    assert result is True


@pytest.mark.asyncio
async def test_named_item_editor_with_line(
    diagram, element_factory, view, event_manager
):
    item = diagram.create(
        UML.classes.DependencyItem, subject=element_factory.create(UML.Dependency)
    )
    view.selection.hovered_item = item
    await view.update()
    result = named_item_editor(item, view, event_manager)

    assert result is True


def test_named_item_editor_without_item(diagram, element_factory, view, event_manager):
    item = diagram.create(UML.classes.DependencyItem)
    result = named_item_editor(item, view, event_manager)

    assert result is False



#
# Configuration file for the Sphinx documentation builder.
#
# This file does only contain a selection of the most common options. For a
# full list see the documentation:
# http://www.sphinx-doc.org/en/master/config

import os

import gi

gi.require_version("Pango", "1.0")
gi.require_version("PangoCairo", "1.0")

from gaphor.application import distribution  # noqa: E402

# -- Project information -----------------------------------------------------

project = "Gaphor"
copyright = "2001, The Gaphor Development Team"
author = "Arjan Molenaar"
version = distribution().version

# -- General configuration ---------------------------------------------------

# If your documentation needs a minimal Sphinx version, state it here.
#
# needs_sphinx = '1.0'

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.doctest",
    "sphinx.ext.intersphinx",
    "sphinx.ext.coverage",
    "sphinx.ext.viewcode",
    "sphinx.ext.autosectionlabel",
    "myst_nb",
    "sphinx_copybutton",
    "gaphor.extensions.sphinx",
    "IPython.sphinxext.ipython_console_highlighting",
]

autodoc_mock_imports = [
    "gi.repository.Gdk",
    "gi.repository.Gtk",
    "gi.repository.Adw",
    "gi.repository.GtkSource",
    "gi.repository.Graphene",
]
gi.require_version = lambda *_: ...

os.environ["PYDEVD_DISABLE_FILE_VALIDATION"] = "1"

# Prefix document path to section labels, otherwise autogenerated labels would look like 'heading'
# rather than 'path/to/file:heading'
autosectionlabel_prefix_document = True

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# The suffix(es) of source filenames.
# You can specify multiple suffix as a list of string:
source_suffix = {".rst": "restructuredtext", ".md": "restructuredtext"}

# The master toctree document.
master_doc = "index"

# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
locale_dirs = ["locale/"]
gettext_compact = "docs"

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = None

nb_execution_show_tb = True
nb_execution_raise_on_error = True
nb_render_markdown_format = "myst"

myst_heading_anchors = 3
suppress_warnings = ["myst.header"]

gaphor_models = {
    "connect": "connect.gaphor",
    "design-principles": "design-principles.gaphor",
    "stereotypes": "stereotypes.gaphor",
    "example": "../examples/sequence-diagram.gaphor",
    "coffee-machine": "../examples/coffee-machine.gaphor",
    "sysml-template": "../gaphor/templates/sysml.gaphor",
    "style-sheets": "style-sheet-examples.gaphor",
    "core": "../models/Core.gaphor",
    "uml": "../models/UML.gaphor",
    "sysml": "../models/SysML.gaphor",
    "raaml": "../models/RAAML.gaphor",
    "c4model": "../models/C4Model.gaphor",
}

# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = "furo"
html_favicon = "../data/logos/gaphor-24x24.png"

# Theme options are theme-specific and customize the look and feel of a theme
# further.  For a list of options available for each theme, see the
# documentation.
#
html_theme_options = {
    "light_logo": "gaphor-logo-full-black.svg",
    "dark_logo": "gaphor-logo-full-white.svg",
    "sidebar_hide_name": True,
}

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = [
    "_static",
]

html_css_files = [
    "css/custom.css",
]
html_js_files = [
    "js/utterances.js",
]

# Custom sidebar templates, must be a dictionary that maps document names
# to template names.
#
# The default sidebars (for documents that don't match any pattern) are
# defined by theme itself.  Builtin themes are using these templates by
# default: ``['localtoc.html', 'relations.html', 'sourcelink.html',
# 'searchbox.html']``.
#
# html_sidebars = {}


# -- Options for HTMLHelp output ---------------------------------------------

# Output file base name for HTML help builder.
htmlhelp_basename = "Gaphordoc"


# -- Options for LaTeX output ------------------------------------------------

latex_elements: dict[str, str] = {
    # The paper size ('letterpaper' or 'a4paper').
    #
    # 'papersize': 'letterpaper',
    # The font size ('10pt', '11pt' or '12pt').
    #
    # 'pointsize': '10pt',
    # Additional stuff for the LaTeX preamble.
    #
    # 'preamble': '',
    # Latex figure (float) alignment
    #
    # 'figure_align': 'htbp',
}

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title,
#  author, documentclass [howto, manual, or own class]).
latex_documents = [(master_doc, "Gaphor.tex", "Gaphor Documentation", author, "manual")]


# -- Options for manual page output ------------------------------------------

# One entry per manual page. List of tuples
# (source start file, name, description, authors, manual section).
man_pages = [(master_doc, "gaphor", "Gaphor Documentation", [author], 1)]


# -- Options for Texinfo output ----------------------------------------------

# Grouping the document tree into Texinfo files. List of tuples
# (source start file, target name, title, author,
#  dir menu entry, description, category)
texinfo_documents = [
    (
        master_doc,
        "Gaphor",
        "Gaphor Documentation",
        author,
        "Gaphor",
        "The simple modeling tool",
        "Miscellaneous",
    )
]


# -- Options for Epub output -------------------------------------------------

# Bibliographic Dublin Core info.
epub_title = project

# The unique identifier of the text. This can be a ISBN number
# or the project homepage.
#
# epub_identifier = ''

# A unique identification for the text.
#
# epub_uid = ''

# A list of files that should not be packed into the epub file.
epub_exclude_files = ["search.html"]


# -- Extension configuration -------------------------------------------------

# -- Options for intersphinx extension ---------------------------------------

# Example configuration for intersphinx: refer to the Python standard library.
intersphinx_mapping = {
    "gnome": ("https://api.pygobject.gnome.org", None),
    "pygobject": ("https://pygobject.gnome.org", None),
    "gaphas": ("https://gaphas.readthedocs.io/en/stable", None),
    "python": ("https://docs.python.org/3", None),
}

# Canonical URL required for custom domain with RTD Addons enabled
html_baseurl = os.environ.get("READTHEDOCS_CANONICAL_URL", "")



from gaphor.core.styling import CompiledStyleSheet
from gaphor.core.styling.inherit import compute_inherited_style
from gaphor.core.styling.tests.test_compiler import Node
from gaphor.diagram.shapes import Box, CssNode


def test_inherit_from_parent_style():
    css = "node { font-size: 10 } node sub { content: 'Hi' }"

    compiled_style_sheet = CompiledStyleSheet(css)
    style = compiled_style_sheet.compute_style(Node("node"))
    inherited = compute_inherited_style(
        style, CssNode("sub", None, Box()).style_node(style)
    )

    assert inherited.get("font-size") == 10
    assert inherited.get("content") == "Hi"


def test_should_not_inherit_everything():
    css = """
    node { background-color: blue }
    node sub { content: 'Hi' }
    node sub * { font-size: 10 }
    """

    compiled_style_sheet = CompiledStyleSheet(css)
    style = compiled_style_sheet.compute_style(Node("node"))
    inherited = compute_inherited_style(
        style,
        CssNode("sub", None, Box()).style_node(
            compute_inherited_style(
                style, CssNode("sub", None, Box()).style_node(style)
            )
        ),
    )

    assert not inherited.get("background-color")
    assert inherited.get("content") == "Hi"
    assert inherited.get("font-size") == 10



import os

import pytest
import pytest_asyncio
from gi.repository import Gdk

from gaphor import UML
from gaphor.core.modeling import Diagram
from gaphor.diagram.general import Box
from gaphor.ui.diagrampage import (
    DiagramPage,
    delete_selected_items,
    get_placement_cursor,
    placement_icon_base,
)
from gaphor.UML import Comment
from gaphor.UML.diagramitems import ClassItem, PackageItem
from gaphor.UML.general.comment import CommentItem


@pytest_asyncio.fixture
async def page(diagram, event_manager, element_factory, modeling_language):
    page = DiagramPage(diagram, event_manager, element_factory, modeling_language)
    page.construct()
    assert page.diagram == diagram
    assert page.view.model == diagram
    yield page
    page.close()


def test_creation(page, element_factory):
    assert len(element_factory.lselect()) == 1
    assert len(element_factory.lselect(Diagram)) == 1


def test_placement(diagram, page, element_factory):
    box = diagram.create(Box)
    page.view.request_update([box])

    diagram.create(CommentItem, subject=element_factory.create(Comment))
    assert len(element_factory.lselect()) == 4


@pytest.mark.skipif(
    bool(os.environ.get("GDK_PIXBUF_MODULEDIR")),
    reason="Causes a SegFault when run from VSCode",
)
def test_placement_icon_base_is_loaded_once():
    icon1 = placement_icon_base()
    icon2 = placement_icon_base()

    assert icon1 is icon2


@pytest.mark.skipif(
    bool(os.environ.get("GDK_PIXBUF_MODULEDIR")),
    reason="Causes a SegFault when run from VSCode",
)
def test_placement_cursor():
    display = Gdk.Display.get_default()
    cursor = get_placement_cursor(display, "gaphor-box-symbolic")

    assert cursor


@pytest.mark.asyncio
async def test_delete_selected_items(create, diagram, view, event_manager):
    package_item = create(PackageItem, UML.Package)
    view.selection.select_items(package_item)
    await view.update()

    delete_selected_items(view, event_manager)

    assert not diagram.ownedPresentation


def test_delete_selected_owner(create, diagram, view, event_manager, sanitizer_service):
    class_item = create(ClassItem, UML.Class)
    diagram.element = class_item.subject
    view.selection.select_items(class_item)

    delete_selected_items(view, event_manager)

    assert not diagram.ownedPresentation
    assert diagram.element is None


@pytest.mark.asyncio
async def test_not_delete_selected_package_owner(
    create, diagram, view, event_manager, sanitizer_service
):
    package_item = create(PackageItem, UML.Package)
    package = package_item.subject
    diagram.element = package_item.subject
    view.selection.select_items(package_item)
    await view.update()

    delete_selected_items(view, event_manager)

    assert not diagram.ownedPresentation
    assert diagram.element is package



import io

import pytest
from dulwich.repo import Repo

from gaphor.storage.mergeconflict import split_ours_and_theirs
from gaphor.storage.tests.fixtures import create_merge_conflict


@pytest.mark.filterwarnings("ignore:use .* Repo._get_user_identity:DeprecationWarning")
def test_split_git_repo(tmp_path):
    repo = Repo.init(tmp_path)
    filename = "testfile.txt"
    test_file = tmp_path / filename

    create_merge_conflict(
        repo,
        test_file,
        initial_text="Initial commit",
        our_text="Second commit",
        their_text="Branch commit",
    )

    ancestor = io.BytesIO()
    ours = io.BytesIO()
    theirs = io.BytesIO()
    result = split_ours_and_theirs(test_file, ancestor, ours, theirs)

    assert result
    assert ancestor.getbuffer() == b"Initial commit"
    assert ours.getbuffer() == b"Second commit"
    assert theirs.getbuffer() == b"Branch commit"



from gaphor.diagram.tests.fixtures import connect
from gaphor.UML.classes import AssociationItem, ClassItem
from gaphor.UML.uml import Actor, Class, UseCase
from gaphor.UML.usecases.actor import ActorItem
from gaphor.UML.usecases.usecase import UseCaseItem


def test_connect_usecase_actor(create):
    actor_item = create(ActorItem, element_class=Actor)
    use_case_item = create(UseCaseItem, element_class=UseCase)
    association_item = create(AssociationItem)

    connect(association_item, association_item.head, actor_item)
    connect(association_item, association_item.tail, use_case_item)

    assert association_item.subject
    assert actor_item.subject in association_item.subject.ownedMember[:].type
    assert use_case_item.subject in association_item.subject.ownedMember[:].type


def test_connect_class_actor(create):
    actor_item = create(ActorItem, element_class=Actor)
    class_item = create(ClassItem, element_class=Class)
    association_item = create(AssociationItem)

    connect(association_item, association_item.head, actor_item)
    connect(association_item, association_item.tail, class_item)

    assert association_item.subject
    assert actor_item.subject in association_item.subject.ownedMember[:].type
    assert class_item.subject in association_item.subject.ownedMember[:].type



import pytest

from gaphor.application import Session
from gaphor.diagram.event import DiagramOpened
from gaphor.ui.abc import UIComponent
from gaphor.ui.event import ElementOpened
from gaphor.ui.tests.fixtures import iterate_until
from gaphor.UML import Comment, Diagram


@pytest.fixture
def session():
    session = Session(
        services=[
            "diagrams",
            "element_editor",
            "event_manager",
            "export_menu",
            "component_registry",
            "element_factory",
            "main_window",
            "model_browser",
            "model_changed",
            "modeling_language",
            "properties",
            "toolbox",
            "tools_menu",
        ]
    )
    main_w = session.get_service("main_window")
    main_w.open()
    yield session
    session.shutdown()


def get_current_diagram(session):
    return (
        session.get_service("component_registry")
        .get(UIComponent, "diagrams")
        .get_current_diagram()
    )


def test_creation(session):
    assert get_current_diagram(session) is None


@pytest.mark.asyncio
async def test_show_diagram(session):
    element_factory = session.get_service("element_factory")
    diagram = element_factory.create(Diagram)

    event_manager = session.get_service("event_manager")
    event_manager.handle(DiagramOpened(diagram))
    assert get_current_diagram(session) == diagram


@pytest.mark.asyncio
async def test_close_diagram(session):
    element_factory = session.get_service("element_factory")
    diagram = element_factory.create(Diagram)

    event_manager = session.get_service("event_manager")
    event_manager.handle(DiagramOpened(diagram))

    diagrams = session.get_service("diagrams")
    diagrams.close_current_tab()

    assert not get_current_diagram(session)


@pytest.mark.asyncio
async def test_open_element_on_diagram(session):
    element_factory = session.get_service("element_factory")
    diagram = element_factory.create(Diagram)
    comment = element_factory.create(Comment)

    event_manager = session.get_service("event_manager")
    event_manager.handle(DiagramOpened(diagram))

    event_manager.handle(ElementOpened(comment))

    assert comment.presentation
    assert comment.presentation[0] in get_current_diagram(session).ownedPresentation


@pytest.mark.asyncio
async def test_update_diagram_name(session):
    element_factory = session.get_service("element_factory")
    diagram = element_factory.create(Diagram)

    event_manager = session.get_service("event_manager")
    event_manager.handle(DiagramOpened(diagram))

    diagram.name = "Foo"


@pytest.mark.asyncio
async def test_flush_model(session):
    element_factory = session.get_service("element_factory")
    diagram = element_factory.create(Diagram)

    event_manager = session.get_service("event_manager")
    event_manager.handle(DiagramOpened(diagram))

    element_factory.flush()


@pytest.mark.skip(reason="May cause funky window manager behavior")
def test_window_mode_maximized(session):
    main_window = session.get_service("main_window")
    properties = session.get_service("properties")

    main_window.window.unfullscreen()
    main_window.window.maximize()
    iterate_until(lambda: properties.get("ui.window-mode") == "maximized")

    assert properties.get("ui.window-mode") == "maximized"

    main_window.window.unmaximize()


@pytest.mark.skip(reason="May cause funky window manager behavior")
def test_window_mode_fullscreened(session):
    main_window = session.get_service("main_window")
    properties = session.get_service("properties")

    main_window.window.fullscreen()
    iterate_until(lambda: properties.get("ui.window-mode") == "fullscreened")

    assert properties.get("ui.window-mode") == "fullscreened"

    main_window.window.unfullscreen()


def test_window_mode_normal(session):
    main_window = session.get_service("main_window")
    properties = session.get_service("properties")

    main_window.window.unfullscreen()
    main_window.window.unmaximize()
    iterate_until(lambda: properties.get("ui.window-mode") == "")

    assert properties.get("ui.window-mode") == ""



"""Test connector item connectors."""

from gaphor import UML
from gaphor.diagram.tests.fixtures import allow, connect, disconnect
from gaphor.UML.classes.component import ComponentItem
from gaphor.UML.classes.dependency import DependencyItem
from gaphor.UML.classes.interface import Folded, InterfaceItem, Side
from gaphor.UML.classes.interfacerealization import InterfaceRealizationItem
from gaphor.UML.deployments import ConnectorItem
from gaphor.UML.deployments.connectorconnect import LegacyConnectorConnectBase


def provide(component, interface):
    """Change component's data so it implements interfaces."""
    impl = component.model.create(UML.InterfaceRealization)
    component.interfaceRealization = impl
    impl.contract = interface


def require(component, interface):
    """Change component's data so it requires interface."""
    usage = component.model.create(UML.Usage)
    component.clientDependency = usage
    usage.supplier = interface


def test_glue(create):
    """Test gluing connector to component."""

    component = create(ComponentItem, UML.Component)
    line = create(ConnectorItem)

    glued = allow(line, line.head, component)
    assert glued


def test_component_connection(create):
    """Test connecting connector to a component."""
    component = create(ComponentItem, UML.Component)
    line = create(ConnectorItem)

    connect(line, line.head, component)
    assert line.subject is None


def test_glue_both(create):
    """Test gluing connector to component when one is connected."""

    c1 = create(ComponentItem, UML.Component)
    c2 = create(ComponentItem, UML.Component)
    line = create(ConnectorItem)

    connect(line, line.head, c1)
    glued = allow(line, line.tail, c2)
    assert not glued


def test_non_folded_glue(create):
    """Test non-folded interface gluing."""

    iface = create(InterfaceItem, UML.Component)
    line = create(ConnectorItem)

    glued = allow(line, line.head, iface)
    assert glued


def test_folded_glue(create):
    """Test folded interface gluing."""

    iface = create(InterfaceItem, UML.Component)
    line = create(ConnectorItem)

    iface.folded = Folded.REQUIRED
    glued = allow(line, line.head, iface)
    assert glued


def test_glue_when_dependency_connected(create):
    """Test interface gluing, when dependency connected."""

    iface = create(InterfaceItem, UML.Component)
    dep = create(DependencyItem)
    line = create(ConnectorItem)

    connect(dep, dep.head, iface)

    iface.folded = Folded.REQUIRED
    glued = allow(line, line.head, iface)
    assert glued


def test_glue_when_implementation_connected(create):
    """Test interface gluing, when implementation connected."""

    iface = create(InterfaceItem, UML.Component)
    impl = create(InterfaceRealizationItem)
    line = create(ConnectorItem)

    connect(impl, impl.head, iface)

    iface.folded = Folded.REQUIRED
    glued = allow(line, line.head, iface)
    assert glued


def test_glue_when_connector_connected(create):
    """Test interface gluing, when connector connected."""

    iface = create(InterfaceItem, UML.Interface)
    comp = create(ComponentItem, UML.Component)
    iface.folded = Folded.REQUIRED

    line1 = create(ConnectorItem)
    line2 = create(ConnectorItem)

    connect(line1, line1.tail, comp)
    connect(line1, line1.head, iface)
    assert Folded.PROVIDED == iface.folded

    glued = allow(line2, line2.head, iface)
    assert glued


def test_simple_connection(create, diagram):
    """Test simple connection to an interface."""
    iface = create(InterfaceItem, UML.Interface)
    comp = create(ComponentItem, UML.Component)
    line = create(ConnectorItem)

    connect(line, line.head, iface)
    connect(line, line.tail, comp)
    iface.update_shapes()
    diagram.update({iface, comp, line})

    # interface goes into assembly mode
    assert iface.folded == Folded.PROVIDED
    # no UML metamodel yet
    assert line.subject


def test_simple_disconnection(create):
    """Test disconnection of simple connection to an interface."""
    iface = create(InterfaceItem, UML.Component)
    line = create(ConnectorItem)

    iface.folded = Folded.PROVIDED

    connect(line, line.head, iface)

    disconnect(line, line.head)
    assert Folded.PROVIDED == iface.folded
    assert iface.side == Side.N

    assert all(p.connectable for p in iface.ports())


def test_getting_component(create):
    """Test getting component."""
    conn1 = create(ConnectorItem)
    conn2 = create(ConnectorItem)

    c1 = create(ComponentItem, UML.Component)
    c2 = create(ComponentItem, UML.Component)

    # connect component
    connect(conn1, conn1.tail, c1)
    connect(conn2, conn2.head, c2)

    assert c1 is LegacyConnectorConnectBase.get_component(conn1)
    assert c2 is LegacyConnectorConnectBase.get_component(conn2)


def test_connection(create):
    """Test basic assembly connection."""
    conn1 = create(ConnectorItem)
    conn2 = create(ConnectorItem)

    c1 = create(ComponentItem, UML.Component)
    c2 = create(ComponentItem, UML.Component)

    iface = create(InterfaceItem, UML.Interface)
    iface.folded = Folded.ASSEMBLY

    # first component provides interface
    # and the second one requires it
    provide(c1.subject, iface.subject)
    require(c2.subject, iface.subject)

    # connect component
    connect(conn1, conn1.head, c1)
    connect(conn2, conn2.head, c2)

    # make an assembly
    connect(conn1, conn1.tail, iface)
    connect(conn2, conn2.tail, iface)

    # test UML data model
    assert conn1.subject is conn2.subject, f"{conn1.subject} is not {conn2.subject}"
    assembly = conn1.subject
    assert isinstance(assembly, UML.Connector)
    assert "assembly" == assembly.kind

    # there should be two connector ends
    assert len(assembly.end) == 2


def test_required_port_glue(create):
    """Test if required port gluing works."""

    conn1 = create(ConnectorItem)
    conn2 = create(ConnectorItem)

    c1 = create(ComponentItem, UML.Component)
    c2 = create(ComponentItem, UML.Component)

    iface = create(InterfaceItem, UML.Interface)
    iface.folded = Folded.ASSEMBLY
    rport = iface.ports()[2]

    provide(c1.subject, iface.subject)
    require(c2.subject, iface.subject)

    # connect components
    connect(conn1, conn1.head, c1)
    connect(conn2, conn2.head, c2)

    connect(conn1, conn1.tail, iface)
    glued = allow(conn2, conn2.tail, iface, rport)
    assert glued


def test_connection_order(create):
    """Test connection order of assembly connection."""
    conn1 = create(ConnectorItem)
    conn2 = create(ConnectorItem)

    c1 = create(ComponentItem, UML.Component)
    c2 = create(ComponentItem, UML.Component)

    iface = create(InterfaceItem, UML.Interface)

    # both components provide interface only
    provide(c1.subject, iface.subject)
    provide(c2.subject, iface.subject)

    # connect components
    connect(conn1, conn1.head, c1)
    connect(conn2, conn2.head, c2)

    # connect to provided port
    connect(conn1, conn1.tail, iface)
    connect(conn2, conn2.tail, iface)
    # no UML data model yet (no connection on required port)
    assert conn1.subject
    assert conn2.subject


def test_additional_connections(create):
    """Test additional connections to assembly connection."""
    conn1 = create(ConnectorItem)
    conn2 = create(ConnectorItem)
    conn3 = create(ConnectorItem)

    c1 = create(ComponentItem, UML.Component)
    c2 = create(ComponentItem, UML.Component)
    c3 = create(ComponentItem, UML.Component)

    iface = create(InterfaceItem, UML.Interface)
    iface.folded = Folded.ASSEMBLY

    # provide and require interface by components
    provide(c1.subject, iface.subject)
    require(c2.subject, iface.subject)
    require(c3.subject, iface.subject)

    # connect components
    connect(conn1, conn1.head, c1)
    connect(conn2, conn2.head, c2)
    connect(conn3, conn3.head, c3)

    # create assembly
    connect(conn1, conn1.tail, iface)
    connect(conn2, conn2.tail, iface)

    # test precondition
    assert conn1.subject and conn2.subject

    #  additional connection
    connect(conn3, conn3.tail, iface)

    # test UML data model
    assert conn3.subject is conn1.subject

    assembly = conn1.subject

    assert 3 == len(assembly.end)


def test_disconnection(create, element_factory):
    """Test assembly connector disconnection."""
    conn1 = create(ConnectorItem)
    conn2 = create(ConnectorItem)

    c1 = create(ComponentItem, UML.Component)
    c2 = create(ComponentItem, UML.Component)

    iface = create(InterfaceItem, UML.Interface)
    iface.folded = Folded.ASSEMBLY

    # first component provides interface
    # and the second one requires it
    provide(c1.subject, iface.subject)
    require(c2.subject, iface.subject)

    # connect component
    connect(conn1, conn1.head, c1)
    connect(conn2, conn2.head, c2)

    # make an assembly
    connect(conn1, conn1.tail, iface)
    connect(conn2, conn2.tail, iface)

    # test precondition
    assert conn1.subject is conn2.subject

    disconnect(conn1, conn1.head)

    assert conn1.subject is None
    assert conn2.subject is None

    assert len(element_factory.lselect(UML.Connector)) == 0
    assert len(element_factory.lselect(UML.ConnectorEnd)) == 0
    assert len(element_factory.lselect(UML.Port)) == 0



import pytest

from gaphor import UML
from gaphor.diagram.connectors import Connector
from gaphor.diagram.copypaste import copy_full, paste_link
from gaphor.UML.actions.action import ActionItem
from gaphor.UML.actions.pin import InputPinItem, OutputPinItem
from gaphor.UML.actions.pinconnect import ActionPinConnector


@pytest.fixture
def action_item(diagram, element_factory):
    return diagram.create(ActionItem, subject=element_factory.create(UML.Action))


@pytest.fixture
def input_pin_item(diagram):
    return diagram.create(InputPinItem)


@pytest.fixture
def output_pin_item(diagram):
    return diagram.create(OutputPinItem)


def test_input_pin_can_connect_to_action(action_item, input_pin_item):
    connector = Connector(action_item, input_pin_item)

    assert isinstance(connector, ActionPinConnector)
    assert connector.allow(input_pin_item.handles()[0], action_item.ports()[0])


def test_output_pin_can_connect_to_action(action_item, output_pin_item):
    connector = Connector(action_item, output_pin_item)

    assert isinstance(connector, ActionPinConnector)
    assert connector.allow(output_pin_item.handles()[0], action_item.ports()[0])


def test_connect_input_pin_to_action(action_item, input_pin_item):
    connector = Connector(action_item, input_pin_item)

    connected = connector.connect(input_pin_item.handles()[0], action_item.ports()[0])

    assert connected
    assert input_pin_item.subject
    assert input_pin_item.subject in action_item.subject.inputValue


def test_disconnect_input_pin_to_action(action_item, input_pin_item):
    connector = Connector(action_item, input_pin_item)
    connector.connect(input_pin_item.handles()[0], action_item.ports()[0])

    connector.disconnect(input_pin_item.handles()[0])

    assert input_pin_item.subject is None
    assert input_pin_item.diagram


def test_connect_output_pin_to_action(action_item, output_pin_item):
    connector = Connector(action_item, output_pin_item)

    connected = connector.connect(output_pin_item.handles()[0], action_item.ports()[0])

    assert connected
    assert output_pin_item.subject
    assert output_pin_item.subject in action_item.subject.outputValue


def test_disconnect_output_pin_to_action(action_item, output_pin_item):
    connector = Connector(action_item, output_pin_item)
    connector.connect(output_pin_item.handles()[0], action_item.ports()[0])

    connector.disconnect(output_pin_item.handles()[0])

    assert output_pin_item.subject is None
    assert output_pin_item.diagram


def test_delete_action_leaves_pins_in_place(
    diagram, action_item, input_pin_item, element_factory
):
    connector = Connector(action_item, input_pin_item)
    connector.connect(input_pin_item.handles()[0], action_item.ports()[0])

    copy_data = copy_full({action_item, input_pin_item})

    new_elements = paste_link(copy_data, diagram)

    new_action_item = next(e for e in new_elements if isinstance(e, ActionItem))
    new_input_pin_item = next(e for e in new_elements if isinstance(e, InputPinItem))

    new_action_item.unlink()

    assert new_action_item.diagram is None
    assert not new_action_item.children

    assert action_item in element_factory
    assert input_pin_item in element_factory
    assert new_action_item not in element_factory
    assert new_input_pin_item not in element_factory



"""Test call behaviour actions."""

from gaphor import UML
from gaphor.UML.actions.action import CallBehaviorActionItem


def test_create_callbehavioraction(diagram, element_factory):
    action = element_factory.create(UML.CallBehaviorAction)
    item = diagram.create(CallBehaviorActionItem, subject=action)

    assert item.subject is action


def test_load_action(element_factory, diagram, saver, loader):
    action = diagram.create(CallBehaviorActionItem)

    action.handles()[0].pos = (1, 2)
    action.matrix.translate(10, 20)
    action.width = 201
    action.height = 202

    dump = saver()
    loader(dump)

    new_action = next(element_factory.select(CallBehaviorActionItem))

    assert tuple(new_action.handles()[0].pos) == (1, 2)
    assert tuple(new_action.matrix) == (1.0, 0.0, 0.0, 1.0, 10, 20)
    assert new_action.width == 201
    assert new_action.height == 202



files = ["dist/Gaphor.app"]
symlinks = {"Applications": "/Applications"}
hide_extensions = ["Gaphor.app"]

window_rect = ((200, 120), (700, 400))
background = "_packaging/macos/background.png"

icon_size = 100
icon_locations = {
    "Gaphor.app": (200, 240),
    "Applications": (500, 240),
}



import pytest

from gaphor.C4Model.modelinglanguage import C4ModelLanguage
from gaphor.core.modeling.modelinglanguage import (
    CoreModelingLanguage,
    MockModelingLanguage,
)
from gaphor.UML.modelinglanguage import UMLModelingLanguage


@pytest.fixture
def modeling_language():
    return MockModelingLanguage(
        CoreModelingLanguage(), UMLModelingLanguage(), C4ModelLanguage()
    )



"""Extension item connection adapter tests."""

from gaphor import UML
from gaphor.core.modeling.diagram import Diagram
from gaphor.diagram.tests.fixtures import allow, connect
from gaphor.UML.classes.klass import ClassItem
from gaphor.UML.profiles.extension import ExtensionItem


def test_glue_head_to_stereotype(element_factory, diagram):
    ext = diagram.create(ExtensionItem)
    st = diagram.create(ClassItem, subject=element_factory.create(UML.Stereotype))

    glued = allow(ext, ext.head, st)

    assert glued


def test_glue_tail_to_stereotype(element_factory, diagram):
    ext = diagram.create(ExtensionItem)
    st = diagram.create(ClassItem, subject=element_factory.create(UML.Stereotype))

    glued = allow(ext, ext.tail, st)

    assert glued


def test_glue_head_to_class(element_factory, diagram):
    ext = diagram.create(ExtensionItem)
    cls = diagram.create(ClassItem, subject=element_factory.create(UML.Class))

    glued = allow(ext, ext.head, cls)

    assert glued


def test_glue_tail_to_class(element_factory, diagram):
    ext = diagram.create(ExtensionItem)
    cls = diagram.create(ClassItem, subject=element_factory.create(UML.Class))

    glued = allow(ext, ext.tail, cls)

    assert not glued


def test_connection(element_factory, diagram):
    ext = diagram.create(ExtensionItem)
    st = diagram.create(ClassItem, subject=element_factory.create(UML.Stereotype))
    cls = diagram.create(ClassItem, subject=element_factory.create(UML.Class))

    connect(ext, ext.tail, st)
    connect(ext, ext.head, cls)

    assert ext.subject


def test_connection_namespace(element_factory, diagram):
    pkg = element_factory.create(UML.Package)
    diagram.element = pkg
    ext = diagram.create(ExtensionItem)
    st = diagram.create(ClassItem, subject=element_factory.create(UML.Stereotype))
    cls = diagram.create(ClassItem, subject=element_factory.create(UML.Class))

    connect(ext, ext.tail, st)
    connect(ext, ext.head, cls)

    assert ext.subject.package is pkg


def test_reuse_extension_in_new_diagram(element_factory, diagram):
    ext = diagram.create(ExtensionItem)
    st = diagram.create(ClassItem, subject=element_factory.create(UML.Stereotype))
    cls = diagram.create(ClassItem, subject=element_factory.create(UML.Class))

    connect(ext, ext.tail, st)
    connect(ext, ext.head, cls)

    diagram2 = element_factory.create(Diagram)
    ext2 = diagram2.create(ExtensionItem)
    st2 = diagram2.create(ClassItem, subject=st.subject)
    cls2 = diagram2.create(ClassItem, subject=cls.subject)

    connect(ext2, ext2.tail, st2)
    connect(ext2, ext2.head, cls2)

    assert ext.subject is ext2.subject



# ruff: noqa: F401

import os

from hypothesis import Phase, settings

from gaphor.conftest import (
    assert_not_in_transaction,
    create,
    diagram,
    element_factory,
    event_manager,
    modeling_language,
    test_models,
    tmp_get_cache_config_dir,
)

settings.register_profile(
    "test",
    derandomize=True,
    max_examples=10,
    stateful_step_count=100,
    deadline=20000,
    phases=[Phase.generate],
)
settings.register_profile("ci", max_examples=1000)
settings.load_profile("test")



import pytest
import pytest_asyncio
from gi.repository import Gtk

from gaphor import UML
from gaphor.C4Model.toolbox import c4model_toolbox_actions
from gaphor.core.eventmanager import EventManager
from gaphor.core.modeling.elementfactory import ElementFactory
from gaphor.diagram.diagramtoolbox import get_tool_def
from gaphor.diagram.tools.placement import PlacementState, on_drag_begin
from gaphor.RAAML.toolbox import raaml_toolbox_actions
from gaphor.SysML.toolbox import sysml_toolbox_actions
from gaphor.ui.diagrampage import DiagramPage
from gaphor.UML.modelinglanguage import UMLModelingLanguage
from gaphor.UML.toolbox import uml_toolbox_actions


@pytest.fixture
def event_manager():
    return EventManager()


@pytest.fixture
def element_factory(event_manager):
    element_factory = ElementFactory(event_manager)
    yield element_factory
    element_factory.shutdown()


@pytest.fixture
def properties():
    return {}


@pytest_asyncio.fixture()
async def tab(event_manager, element_factory):
    diagram = element_factory.create(UML.Diagram)
    tab = DiagramPage(diagram, event_manager, element_factory, UMLModelingLanguage())

    window = Gtk.Window.new()
    window.set_child(tab.construct())

    window.set_visible(True)
    yield tab
    window.destroy()


def test_pointer(tab):
    tab.apply_tool_set("toolbox-pointer")

    # TODO: what's the observed behavior?
    assert tab.view._controllers  # noqa: SLF001


def flatten(list):
    return [item for sublist in list for item in sublist]


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "tool_def",
    set(
        flatten(
            [
                flatten(section.tools for section in uml_toolbox_actions),
                flatten(section.tools for section in sysml_toolbox_actions),
                flatten(section.tools for section in raaml_toolbox_actions),
                flatten(section.tools for section in c4model_toolbox_actions),
            ]
        )
    ),
)
async def test_placement_action(tool_def, tab, event_manager):
    if not tool_def.item_factory:
        return

    tool = Gtk.GestureDrag.new()
    tab.view.add_controller(tool)

    placement_state = PlacementState(
        tool_def.item_factory, event_manager, tool_def.handle_index
    )
    on_drag_begin(tool, 0, 0, placement_state)
    await tab.view.update()


@pytest.mark.asyncio
async def test_placement_object_node(tab, element_factory, event_manager):
    await test_placement_action(
        get_tool_def(UMLModelingLanguage(), "toolbox-object-node"), tab, event_manager
    )
    assert len(element_factory.lselect(UML.ObjectNode)) == 1


@pytest.mark.asyncio
async def test_placement_partition(tab, element_factory, event_manager):
    await test_placement_action(
        get_tool_def(UMLModelingLanguage(), "toolbox-partition"), tab, event_manager
    )

    assert len(element_factory.lselect(UML.ActivityPartition)) == 2


@pytest.mark.parametrize(
    "toolbox_actions",
    [
        uml_toolbox_actions,
        sysml_toolbox_actions,
        raaml_toolbox_actions,
        c4model_toolbox_actions,
    ],
)
def test_uml_toolbox_actions_shortcut_unique(toolbox_actions):
    shortcuts: dict[str, list[str]] = {}

    for _category, items in toolbox_actions:
        for action_name, _label, _icon_name, shortcut, *_rest in items:
            try:
                shortcuts[shortcut].append(action_name)
            except KeyError:
                shortcuts[shortcut] = [action_name]

    for key, val in list(shortcuts.items()):
        if key is not None:
            assert len(val) == 1, f"Duplicate toolbox shortcut {key} for {val}"



"""Unit tests for AssociationItem."""

import pytest

from gaphor import UML
from gaphor.diagram.tests.fixtures import connect, get_connected
from gaphor.UML.classes.association import (
    AssociationItem,
    draw_default_head,
    draw_head_navigable,
    draw_head_none,
)
from gaphor.UML.classes.klass import ClassItem


class Items:
    def __init__(self, assoc, class1, class2):
        self.assoc = assoc
        self.class1 = class1
        self.class2 = class2


@pytest.fixture
def items(create):
    return Items(
        create(AssociationItem),
        create(ClassItem, UML.Class),
        create(ClassItem, UML.Class),
    )


def test_create(items):
    """Test association creation and its basic properties."""
    connect(items.assoc, items.assoc.head, items.class1)
    connect(items.assoc, items.assoc.tail, items.class2)

    assert isinstance(items.assoc.subject, UML.Association)
    assert items.assoc.head_subject is not None
    assert items.assoc.tail_subject is not None

    assert not items.assoc.show_direction

    items.assoc.show_direction = True
    assert items.assoc.show_direction


def test_direction(items):
    """Test association direction inverting."""
    connect(items.assoc, items.assoc.head, items.class1)
    connect(items.assoc, items.assoc.tail, items.class2)

    assert items.assoc.head_subject is items.assoc.subject.memberEnd[0]
    assert items.assoc.tail_subject is items.assoc.subject.memberEnd[1]


def test_invert_direction(items):
    connect(items.assoc, items.assoc.head, items.class1)
    connect(items.assoc, items.assoc.tail, items.class2)

    items.assoc.invert_direction()

    assert items.assoc.subject.memberEnd
    assert items.assoc.head_subject is items.assoc.subject.memberEnd[1]
    assert items.assoc.tail_subject is items.assoc.subject.memberEnd[0]


def test_association_end_updates(create, diagram):
    """Test association end navigability connected to a class."""
    c1 = create(ClassItem, UML.Class)
    c2 = create(ClassItem, UML.Class)
    a = create(AssociationItem)

    connect(a, a.head, c1)
    c = get_connected(a, a.head)
    assert c is c1

    connect(a, a.tail, c2)
    c = get_connected(a, a.tail)
    assert c is c2

    assert a.subject.memberEnd, a.subject.memberEnd

    assert a.subject.memberEnd[0] is a.head_subject
    assert a.subject.memberEnd[1] is a.tail_subject
    assert a.subject.memberEnd[0].name is None

    a.subject.memberEnd[0].name = "blah"
    diagram.update({a})

    assert a.head_end.name == "+ blah", a.head_end.name


def test_association_end_owner_handles(items):
    assert items.assoc.head_end.owner_handle is items.assoc.head
    assert items.assoc.tail_end.owner_handle is items.assoc.tail


@pytest.mark.parametrize(
    "navigability,draw_func",
    [[None, draw_default_head], [True, draw_head_navigable], [False, draw_head_none]],
)
def test_association_head_end_not_navigable(items, navigability, draw_func):
    connect(items.assoc, items.assoc.head, items.class1)
    connect(items.assoc, items.assoc.tail, items.class2)

    end = items.assoc.head_end
    UML.recipes.set_navigability(end.subject.association, end.subject, navigability)
    items.assoc.update()

    assert items.assoc.head_subject.navigability is navigability
    assert items.assoc.draw_head is draw_func



"""A Property-based test.

Generate a bunch of diagrams and auto-layout them.
"""

from __future__ import annotations

import contextlib
import itertools
from collections.abc import Iterable

import hypothesis
from hypothesis import settings
from hypothesis.control import assume, cleanup
from hypothesis.errors import UnsatisfiedAssumption
from hypothesis.stateful import (
    RuleBasedStateMachine,
    initialize,
    rule,
)
from hypothesis.strategies import data, integers, sampled_from

from gaphor.application import Session
from gaphor.C4Model.toolbox import c4
from gaphor.core import Transaction
from gaphor.core.modeling import ElementFactory, Presentation
from gaphor.core.modeling.base import generate_id, uuid_generator
from gaphor.diagram.group import can_group, change_owner
from gaphor.diagram.presentation import LinePresentation
from gaphor.diagram.tests.fixtures import allow, connect
from gaphor.plugins.autolayout import AutoLayout
from gaphor.RAAML.toolbox import fta, stpa
from gaphor.SysML.toolbox import blocks, internal_blocks, requirements
from gaphor.UML import Diagram, Element, Package, diagramitems
from gaphor.UML.toolbox import (
    actions,
    classes,
    deployments,
    interactions,
    profiles,
    states,
    use_cases,
)


def tooldef():
    return sampled_from(
        list(
            itertools.chain(
                # UML
                classes.tools,
                deployments.tools,
                use_cases.tools,
                profiles.tools,
                actions.tools,
                interactions.tools,
                states.tools,
                # C4
                c4.tools,
                # SysML
                blocks.tools,
                internal_blocks.tools,
                requirements.tools,
                # RAAML
                stpa.tools,
                fta.tools,
            )
        )
    )


def load_default_model(element_factory):
    element_factory.flush()
    with element_factory.block_events():
        model = element_factory.create(Package)
        model.name = "New model"
        diagram = element_factory.create(Diagram)
        diagram.element = model
        diagram.name = "New diagram"


def ordered(elements: Iterable[Element]) -> list[Element]:
    return sorted(elements, key=lambda e: e.id)


def relations(diagram):
    relations = [
        p for p in diagram.presentation if isinstance(p, diagramitems.DependencyItem)
    ]
    assume(relations)
    return sampled_from(ordered(relations))


class AutoLayouting(RuleBasedStateMachine):
    @property
    def model(self) -> ElementFactory:
        return self.session.get_service("element_factory")  # type: ignore[no-any-return]

    @property
    def auto_layout(self) -> AutoLayout:
        return self.session.get_service("auto_layout")  # type: ignore[no-any-return]

    @property
    def transaction(self) -> Transaction:
        return Transaction(self.session.get_service("event_manager"))

    def select(self, expression=None):
        elements = ordered(self.model.select(expression))
        assume(elements)
        return sampled_from(elements)

    def targets(self, relation, handle):
        return self.select(
            lambda e: (
                isinstance(e, diagramitems.ClassItem)
                and e.diagram is relation.diagram
                and allow(relation, handle, e)
            )
        )

    @initialize()
    def new_session(self):
        generate_id(map(str, itertools.count()))
        cleanup(lambda: generate_id(uuid_generator()))

        self.session = Session()
        self.copy_buffer: object = None

        load_default_model(self.model)
        self.diagram = next(self.model.select(Diagram))

    @rule(
        tooldef=tooldef(),
        data=data(),
        x=integers(min_value=0, max_value=2000),
        y=integers(min_value=0, max_value=2000),
    )
    def add_item_to_diagram(self, tooldef, data, x, y):
        with self.transaction:
            item: Presentation = tooldef.item_factory(self.diagram)
            item.matrix.translate(x, y)
            self.diagram.update({item})

        # Do best effort to connect a line, no problem if it fails
        if isinstance(item, LinePresentation):
            self.try_connect_relation(data, item, item.head)
            self.try_connect_relation(data, item, item.tail)

    def try_connect_relation(self, data, item, handle):
        with contextlib.suppress(UnsatisfiedAssumption):
            self._connect_relation(data, item, handle)

    @rule(data=data())
    def connect_relation(self, data):
        relation = data.draw(relations(self.diagram))
        handle = data.draw(sampled_from([relation.head, relation.tail]))
        self._connect_relation(data, relation, handle)

    def _connect_relation(self, data, relation, handle):
        target = data.draw(self.targets(relation, handle))
        with self.transaction:
            connect(relation, handle, target)

    @rule(data=data())
    def change_owner(self, data):
        parent = data.draw(self.select())
        element = data.draw(
            self.select(lambda e: e is not parent and can_group(parent, e))
        )
        with self.transaction as tx:
            changed = change_owner(parent, element)
            if not changed:
                tx.rollback()
        assume(changed)

    def teardown(self):
        self.auto_layout.layout(self.diagram)


AutoLayouting.TestCase.settings = settings(
    max_examples=5,
    stateful_step_count=30,
    deadline=2000,
    phases=[hypothesis.Phase.generate],
)
TestAutoLayouting = AutoLayouting.TestCase



from gaphor.diagram.copypaste import copy_full, paste_full, paste_link
from gaphor.UML import (
    Activity,
    ActivityParameterNode,
    InputPin,
    JoinNode,
    LiteralInteger,
    ObjectNode,
    Parameter,
)
from gaphor.UML.actions.activity import ActivityItem


def test_shallow_copy_activity_with_parameter(diagram, element_factory):
    activity_item = diagram.create(
        ActivityItem, subject=element_factory.create(Activity)
    )
    node = element_factory.create(ActivityParameterNode)
    activity_item.subject.node = node
    node.parameter = element_factory.create(Parameter)

    copy_data = copy_full({activity_item})

    new_elements = paste_link(copy_data, diagram)
    new_activity_item = next(
        (e for e in new_elements if isinstance(e, ActivityItem)), None
    )

    assert new_activity_item, new_elements
    assert new_activity_item.children
    assert new_activity_item.children[0].subject is node


def test_deep_copy_activity_with_parameter(diagram, element_factory):
    activity_item = diagram.create(
        ActivityItem, subject=element_factory.create(Activity)
    )
    node = element_factory.create(ActivityParameterNode)
    activity_item.subject.node = node
    node.parameter = element_factory.create(Parameter)
    node.parameter.name = "Name"

    copy_data = copy_full({activity_item})
    new_elements = paste_full(copy_data, diagram)
    new_activity_item = next(
        (e for e in new_elements if isinstance(e, ActivityItem)), None
    )

    assert new_activity_item, new_elements
    assert new_activity_item.children
    assert new_activity_item.children[0].subject is not node
    assert new_activity_item.children[0].subject.parameter.name == node.parameter.name


def test_do_not_copy_activity_parameter_node(diagram, element_factory):
    activity_item = diagram.create(
        ActivityItem, subject=element_factory.create(Activity)
    )
    node = element_factory.create(ActivityParameterNode)
    activity_item.subject.node = node
    node.parameter = element_factory.create(Parameter)

    copy_data = copy_full(activity_item.children)

    assert activity_item.children
    assert not copy_data.elements


def test_copy_pin(diagram, element_factory):
    pin = element_factory.create(InputPin)
    pin.lowerValue = element_factory.create(LiteralInteger)
    pin.lowerValue.value = 1
    pin.upperValue = element_factory.create(LiteralInteger)
    pin.upperValue.value = 2

    buffer = copy_full({pin})
    paste_full(buffer, diagram)

    new_pin = next(p for p in element_factory.select(InputPin) if p is not pin)

    assert new_pin.lowerValue
    assert new_pin.upperValue
    assert new_pin.lowerValue.value == 1
    assert new_pin.upperValue.value == 2

    assert new_pin.lowerValue is not pin.lowerValue
    assert new_pin.upperValue is not pin.upperValue


def test_copy_object_node(diagram, element_factory):
    node = element_factory.create(ObjectNode)
    node.upperBound = element_factory.create(LiteralInteger)
    node.upperBound.value = 1

    buffer = copy_full({node})
    paste_full(buffer, diagram)

    new_node = next(p for p in element_factory.select(ObjectNode) if p is not node)

    assert new_node.upperBound
    assert new_node.upperBound.value == 1

    assert new_node.upperBound is not node.upperBound


def test_copy_join_node(diagram, element_factory):
    node = element_factory.create(JoinNode)
    node.joinSpec = element_factory.create(LiteralInteger)
    node.joinSpec.value = 1

    buffer = copy_full({node})
    paste_full(buffer, diagram)

    new_node = next(p for p in element_factory.select(JoinNode) if p is not node)

    assert new_node.joinSpec
    assert new_node.joinSpec.value == 1

    assert new_node.joinSpec is not node.joinSpec



# ruff: noqa: SLF001

import pytest

from gaphor.core import event_handler
from gaphor.core.modeling import Base, swap_element_type
from gaphor.core.modeling.event import AssociationUpdated
from gaphor.core.modeling.properties import association, attribute, derivedunion
from gaphor.services.undomanager import NotInTransactionException
from gaphor.tests.raises import raises_exception_group
from gaphor.transaction import Transaction


def test_nested_transactions(event_manager, undo_manager):
    assert not undo_manager._current_transaction

    tx = Transaction(event_manager)

    assert undo_manager._current_transaction

    current = undo_manager._current_transaction
    tx2 = Transaction(event_manager)
    assert undo_manager._current_transaction is current

    tx2.commit()

    assert undo_manager._current_transaction is current

    tx.commit()
    assert undo_manager._current_transaction is None


def test_not_in_transaction(undo_manager):
    def action():
        pass

    with pytest.raises(NotInTransactionException):
        undo_manager.add_undo_action(action)

    undo_manager.begin_transaction()
    undo_manager.add_undo_action(action)
    assert undo_manager._current_transaction
    assert undo_manager.can_undo()
    assert len(undo_manager._current_transaction._actions) == 1


def test_actions(event_manager, undo_manager):
    undone = [0]

    def undo_action(undone=undone):
        undone[0] = 1
        undo_manager.add_undo_action(redo_action)

    def redo_action(undone=undone):
        undone[0] = -1
        undo_manager.add_undo_action(undo_action)

    tx = Transaction(event_manager)
    undo_manager.add_undo_action(undo_action)
    assert undo_manager._current_transaction
    assert undo_manager.can_undo()
    assert len(undo_manager._current_transaction._actions) == 1

    tx.commit()

    undo_manager.undo_transaction()
    assert not undo_manager.can_undo(), undo_manager._undo_stack
    assert undone[0] == 1, undone

    undone[0] = 0

    assert undo_manager.can_redo(), undo_manager._redo_stack

    undo_manager.redo_transaction()
    assert not undo_manager.can_redo()
    assert undo_manager.can_undo()
    assert undone[0] == -1, undone


def test_undo_attribute(element_factory, undo_manager):
    class A(Base):
        attr = attribute("attr", bytes, default="default")

    undo_manager.begin_transaction()
    a = element_factory.create(A)
    undo_manager.commit_transaction()

    assert a.attr == "default", a.attr

    undo_manager.begin_transaction()
    a.attr = "five"
    undo_manager.commit_transaction()

    assert a.attr == "five"

    undo_manager.undo_transaction()

    assert a.attr == "default", a.attr

    undo_manager.redo_transaction()

    assert a.attr == "five"


def test_change_element_type(event_manager, element_factory, undo_manager):
    class A(Base):
        pass

    class B(Base):
        pass

    with Transaction(event_manager):
        a = element_factory.create(A)

    with Transaction(event_manager):
        swap_element_type(a, B)
        assert type(a) is B

    undo_manager.undo_transaction()

    assert type(a) is A


def test_no_value_change_if_not_in_transaction(
    event_manager, element_factory, undo_manager
):
    class A(Base):
        attr = attribute("attr", bytes, default="default")

    with Transaction(event_manager):
        a = element_factory.create(A)

    with raises_exception_group(NotInTransactionException):
        a.attr = "foo"

    assert a.attr == "default"


def test_undo_association_1_x(event_manager, element_factory, undo_manager):
    class A(Base):
        pass

    class B(Base):
        pass

    A.one = association("one", B, 0, 1, opposite="two")
    B.two = association("two", A, 0, 1)

    with Transaction(event_manager):
        a = element_factory.create(A)
        b = element_factory.create(B)
    undo_manager.clear_undo_stack()

    assert a.one is None
    assert b.two is None

    undo_manager.begin_transaction()
    a.one = b

    undo_manager.commit_transaction()
    assert a.one is b
    assert b.two is a
    assert len(undo_manager._undo_stack) == 1
    assert len(undo_manager._undo_stack[0]._actions) == 2, undo_manager._undo_stack[
        0
    ]._actions

    undo_manager.undo_transaction()
    assert a.one is None
    assert b.two is None
    assert undo_manager.can_redo()
    assert len(undo_manager._redo_stack) == 1
    assert len(undo_manager._redo_stack[0]._actions) == 3, undo_manager._redo_stack[
        0
    ]._actions

    undo_manager.redo_transaction()
    assert len(undo_manager._undo_stack) == 1
    assert len(undo_manager._undo_stack[0]._actions) == 2, undo_manager._undo_stack[
        0
    ]._actions
    assert b.two is a
    assert a.one is b


def test_undo_association_1_n(event_manager, element_factory, undo_manager):
    class A(Base):
        pass

    class B(Base):
        pass

    A.one = association("one", B, upper=1, opposite="two")
    B.two = association("two", A, upper="*", opposite="one")

    with Transaction(event_manager):
        a1 = element_factory.create(A)
        a2 = element_factory.create(A)
        b1 = element_factory.create(B)
        element_factory.create(B)

    undo_manager.clear_undo_stack()

    undo_manager.begin_transaction()
    b1.two = a1

    undo_manager.commit_transaction()
    assert a1 in b1.two
    assert b1 is a1.one
    assert len(undo_manager._undo_stack) == 1
    assert len(undo_manager._undo_stack[0]._actions) == 2, undo_manager._undo_stack[
        0
    ]._actions

    undo_manager.undo_transaction()
    assert len(b1.two) == 0
    assert a1.one is None
    assert undo_manager.can_redo()
    assert len(undo_manager._redo_stack) == 1
    assert len(undo_manager._redo_stack[0]._actions) == 2, undo_manager._redo_stack[
        0
    ]._actions

    undo_manager.redo_transaction()
    assert a1 in b1.two
    assert b1 is a1.one

    undo_manager.begin_transaction()
    b1.two = a2

    undo_manager.commit_transaction()
    assert a1 in b1.two
    assert a2 in b1.two
    assert b1 is a1.one
    assert b1 is a2.one


def test_undo_association_undo_in_same_order(
    event_manager, element_factory, undo_manager
):
    class A(Base):
        pass

    class B(Base):
        pass

    A.one = association("one", B, upper=1, opposite="two")
    B.two = association("two", A, upper="*", opposite="one")

    with Transaction(event_manager):
        a1 = element_factory.create(A)
        a2 = element_factory.create(A)
        a3 = element_factory.create(A)
        b = element_factory.create(B)
        b.two = a1
        b.two = a2
        b.two = a3

    with Transaction(event_manager):
        del b.two[a2]

    undo_manager.undo_transaction()

    assert [e.id for e in b.two] == [a1.id, a2.id, a3.id]


def test_element_factory_undo(element_factory, undo_manager):
    undo_manager.begin_transaction()
    p = element_factory.create(Base)

    assert undo_manager._current_transaction
    assert undo_manager._current_transaction._actions
    assert undo_manager.can_undo()

    undo_manager.commit_transaction()
    assert undo_manager.can_undo()
    assert element_factory.size() == 1

    undo_manager.undo_transaction()
    assert not undo_manager.can_undo()
    assert undo_manager.can_redo()
    assert element_factory.size() == 0

    undo_manager.redo_transaction()
    assert undo_manager.can_undo()
    assert not undo_manager.can_redo()
    assert element_factory.size() == 1
    assert element_factory.lookup(p.id)


def test_element_factory_rollback(element_factory, undo_manager):
    undo_manager.begin_transaction()
    element_factory.create(Base)

    assert undo_manager._current_transaction
    assert undo_manager._current_transaction._actions
    assert undo_manager.can_undo()

    undo_manager.rollback_transaction()
    assert not undo_manager.can_undo()
    assert element_factory.size() == 0


def test_uml_associations(event_manager, element_factory, undo_manager):
    class A(Base):
        is_unlinked = False

        def unlink(self):
            self.is_unlinked = True
            Base.unlink(self)

    A.a1 = association("a1", A, upper=1)
    A.a2 = association("a2", A, upper=1)
    A.b1 = association("b1", A, upper="*")
    A.b2 = association("b2", A, upper="*")
    A.b3 = association("b3", A, upper=1)

    A.derived_a = derivedunion("derived_a", A, 0, 1, A.a1, A.a2)
    A.derived_b = derivedunion("derived_b", A, 0, "*", A.b1, A.b2, A.b3)

    events = []

    @event_handler(AssociationUpdated)
    def handler(event, events=events):
        events.append(event)

    event_manager.subscribe(handler)

    undo_manager.begin_transaction()

    a = element_factory.create(A)
    a.a1 = element_factory.create(A)
    undo_manager.commit_transaction()

    assert len(events) == 2, events  # both  AssociationSet and DerivedSet events
    assert events[0].property is A.a1
    assert undo_manager.can_undo()

    undo_manager.undo_transaction()
    assert not undo_manager.can_undo()
    assert undo_manager.can_redo()
    assert len(events) == 4, events
    assert events[2].property is A.a1


def test_set_association_outside_transaction(
    undo_manager, element_factory, event_manager
):
    class A(Base):
        pass

    A.a = association("a", A, upper=1, opposite="b")
    A.b = association("b", A, opposite="a")

    with Transaction(event_manager):
        a = element_factory.create(A)
        b = element_factory.create(A)

    with raises_exception_group(NotInTransactionException):
        a.a = b

    assert not a.b
    assert a.a is None


def test_set_multi_association_outside_transaction(
    undo_manager, element_factory, event_manager
):
    class A(Base):
        pass

    A.a = association("a", A, upper=1, opposite="b")
    A.b = association("b", A, opposite="a")

    with Transaction(event_manager):
        a = element_factory.create(A)
        b = element_factory.create(A)

    with raises_exception_group(NotInTransactionException):
        a.b = b

    assert not a.b
    assert a.a is None


def test_update_association_outside_transaction(
    undo_manager, element_factory, event_manager
):
    class A(Base):
        pass

    A.a = association("a", A, upper=1, opposite="b")
    A.b = association("b", A, opposite="a")

    with Transaction(event_manager):
        a = element_factory.create(A)
        b = element_factory.create(A)
        a.a = b
        other = element_factory.create(A)

    with raises_exception_group(NotInTransactionException):
        a.a = other

    assert a in b.b
    assert a.a is b


def test_set_derived_union_outside_transaction(
    undo_manager, element_factory, event_manager
):
    class A(Base):
        pass

    A.a = association("a", A, upper=1, opposite="b")
    A.b = association("b", A, opposite="a")
    A.derived_a = derivedunion("derived_a", A, 0, 1, A.a)
    A.derived_b = derivedunion("derived_b", A, 0, "*", A.b)

    with Transaction(event_manager):
        a = element_factory.create(A)
        b = element_factory.create(A)

    with raises_exception_group(NotInTransactionException):
        a.a = b

    with raises_exception_group(NotInTransactionException):
        a.b = b

    assert not a.b
    assert not a.a
    assert not a.derived_a
    assert not a.derived_b


def test_rollback_transaction(undo_manager, element_factory, event_manager):
    class A(Base):
        pass

    with Transaction(event_manager) as tx:
        element_factory.create(A)
        tx.rollback()

    assert not undo_manager.can_undo()


def test_redo_stack(event_manager, element_factory, undo_manager):
    undo_manager.begin_transaction()

    p = element_factory.create(Base)

    assert undo_manager._current_transaction
    assert undo_manager._current_transaction._actions
    assert undo_manager.can_undo()

    undo_manager.commit_transaction()
    assert undo_manager.can_undo()
    assert element_factory.size() == 1, element_factory.size()

    with Transaction(event_manager):
        element_factory.create(Base)

    assert undo_manager.can_undo()
    assert not undo_manager.can_redo()
    assert element_factory.size() == 2

    undo_manager.undo_transaction()
    assert undo_manager.can_undo()
    assert 1 == len(undo_manager._undo_stack)
    assert 1 == len(undo_manager._redo_stack)
    assert undo_manager.can_redo()
    assert element_factory.size() == 1

    undo_manager.undo_transaction()
    assert not undo_manager.can_undo()
    assert undo_manager.can_redo()
    assert 0 == len(undo_manager._undo_stack)
    assert 2 == len(undo_manager._redo_stack)

    undo_manager.redo_transaction()
    assert 1 == len(undo_manager._undo_stack)
    assert 1 == len(undo_manager._redo_stack)
    assert undo_manager.can_undo()
    assert undo_manager.can_redo()
    assert element_factory.size() == 1

    undo_manager.redo_transaction()
    assert undo_manager.can_undo()
    assert not undo_manager.can_redo()
    assert element_factory.size() == 2

    assert element_factory.lookup(p.id)


def test_editing_transaction(event_manager, element_factory, undo_manager):
    class A(Base):
        attr = attribute("attr", bytes, default="")

    with Transaction(event_manager):
        a = element_factory.create(A)

    with Transaction(event_manager, context="editing"):
        a.attr = "some text"

    with Transaction(event_manager, context="editing"):
        a.attr = "new text"

    undo_manager.undo_transaction()

    assert a.attr == ""


def test_editing_with_closed_transaction(event_manager, element_factory, undo_manager):
    class A(Base):
        attr = attribute("attr", bytes, default="")

    with Transaction(event_manager):
        a = element_factory.create(A)

    with Transaction(event_manager, context="editing"):
        a.attr = "some text"

    undo_manager.close_transaction()

    with Transaction(event_manager, context="editing"):
        a.attr = "new text"

    undo_manager.undo_transaction()

    assert a.attr == "some text"



from gaphor.diagram.tests.fixtures import find
from gaphor.UML import Comment
from gaphor.UML.general.commentpropertypage import CommentPropertyPage


def test_comment_property_page_body(element_factory, event_manager):
    subject = element_factory.create(Comment)
    property_page = CommentPropertyPage(subject, event_manager)

    widget = property_page.construct()
    comment = find(widget, "comment")
    comment.get_buffer().set_text("test")

    assert subject.body == "test"


def test_comment_property_page_update_text(element_factory, event_manager):
    subject = element_factory.create(Comment)
    property_page = CommentPropertyPage(subject, event_manager)

    widget = property_page.construct()
    comment = find(widget, "comment")
    subject.body = "test"
    buffer = comment.get_buffer()

    assert (
        buffer.get_text(buffer.get_start_iter(), buffer.get_end_iter(), False) == "test"
    )



import pytest

from gaphor.application import Application
from gaphor.core.modeling import Diagram
from gaphor.diagram.general import Line
from gaphor.diagram.segment import Segment
from gaphor.event import SessionShutdown
from gaphor.storage.recovery import sessions_dir, sha256sum
from gaphor.transaction import Transaction
from gaphor.ui import recover_sessions


@pytest.fixture
def application():
    app = Application()
    yield app
    app.shutdown()


@pytest.mark.asyncio
async def test_recovery_when_reloading_file(application: Application, test_models):
    model_file = test_models / "simple-items.gaphor"
    session = application.new_session(filename=model_file)
    element_factory = session.get_service("element_factory")
    with Transaction(session.get_service("event_manager")):
        diagram = element_factory.create(Diagram)

    application.shutdown_session(session)

    new_session = application.recover_session(
        session_id=session.session_id, filename=model_file
    )
    await new_session.get_service("event_manager").gather_tasks()
    new_element_factory = new_session.get_service("element_factory")

    assert new_element_factory.lookup(diagram.id)


@pytest.mark.asyncio
async def test_recovery_when_change_is_rolled_back(
    application: Application, test_models
):
    model_file = test_models / "simple-items.gaphor"
    session = application.new_session(filename=model_file)
    element_factory = session.get_service("element_factory")
    with Transaction(session.get_service("event_manager")) as tx:
        diagram = element_factory.create(Diagram)
        tx.rollback()

    application.shutdown_session(session)

    new_session = application.recover_session(
        session_id=session.session_id, filename=model_file
    )
    await new_session.get_service("event_manager").gather_tasks()
    new_element_factory = new_session.get_service("element_factory")

    assert not new_element_factory.lookup(diagram.id)


@pytest.mark.asyncio
async def test_recovery_when_model_is_loaded_twice(
    application: Application, test_models
):
    model_file = test_models / "simple-items.gaphor"
    session = application.new_session(filename=model_file)
    element_factory = session.get_service("element_factory")
    with Transaction(session.get_service("event_manager")):
        diagram = element_factory.create(Diagram)

    new_session = application.new_session(filename=model_file, force=True)
    new_element_factory = new_session.get_service("element_factory")

    assert not new_element_factory.lookup(diagram.id)


@pytest.mark.asyncio
async def test_no_recovery_for_new_session(application: Application):
    session = application.new_session()
    element_factory = session.get_service("element_factory")
    with Transaction(session.get_service("event_manager")):
        diagram = element_factory.create(Diagram)

    application.shutdown_session(session)

    new_session = application.new_session()
    new_element_factory = new_session.get_service("element_factory")

    assert not new_element_factory.lookup(diagram.id)


@pytest.mark.asyncio
async def test_no_recovery_for_saved_file(
    application: Application, test_models, tmp_path
):
    model_file = test_models / "simple-items.gaphor"
    session = application.new_session(filename=model_file)
    element_factory = session.get_service("element_factory")
    log_file = session.get_service("recovery").event_log.log_file
    with Transaction(session.get_service("event_manager")):
        diagram = element_factory.create(Diagram)

    file_manager = session.get_service("file_manager")
    await file_manager.save(tmp_path / "newfile.gaphor")

    application.shutdown_session(session)

    new_session = application.recover_session(
        session_id=session.session_id, filename=model_file
    )
    await new_session.get_service("event_manager").gather_tasks()
    new_element_factory = new_session.get_service("element_factory")

    assert not new_element_factory.lookup(diagram.id)
    assert not log_file.exists()


@pytest.mark.asyncio
async def test_no_recovey_when_model_changed(
    application: Application, test_models, tmp_path
):
    model_file = tmp_path / "testfile.gaphor"
    model_file.write_text(
        (test_models / "simple-items.gaphor").read_text(encoding="utf-8"),
        encoding="utf-8",
    )

    session = application.new_session(filename=model_file)
    element_factory = session.get_service("element_factory")
    log_file = session.get_service("recovery").event_log.log_file
    with Transaction(session.get_service("event_manager")):
        diagram = element_factory.create(Diagram)

    application.shutdown_session(session)

    # Modify model a bit
    with open(model_file, mode="a", encoding="utf-8") as f:
        f.write("\n")

    new_session = application.recover_session(
        session_id=session.session_id, filename=model_file
    )
    await new_session.get_service("event_manager").gather_tasks()
    new_element_factory = new_session.get_service("element_factory")

    assert not new_element_factory.lookup(diagram.id)
    assert log_file.with_suffix(".recovery.bak").exists()


@pytest.mark.asyncio
async def test_no_recovery_for_properly_closed_session(
    application: Application, test_models
):
    model_file = test_models / "simple-items.gaphor"
    session = application.new_session(filename=model_file)
    event_manager = session.get_service("event_manager")
    element_factory = session.get_service("element_factory")
    with Transaction(event_manager):
        diagram = element_factory.create(Diagram)

    event_manager.handle(SessionShutdown(quitting=False))

    new_session = application.recover_session(
        session_id=session.session_id, filename=model_file
    )
    await new_session.get_service("event_manager").gather_tasks()
    new_element_factory = new_session.get_service("element_factory")

    assert not new_element_factory.lookup(diagram.id)


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "errorous_line",
    [
        '[("s", "1234", "name", "value")]\n',
        '[("s", syntax error)]\n',
        "\n",
    ],
)
async def test_broken_recovery_log(
    application: Application, test_models, caplog, errorous_line
):
    model_file = test_models / "simple-items.gaphor"
    session = application.new_session(filename=model_file)
    element_factory = session.get_service("element_factory")
    log_file = session.get_service("recovery").event_log.log_file
    with Transaction(session.get_service("event_manager")):
        diagram = element_factory.create(Diagram)

    application.shutdown_session(session)

    with log_file.open("a", encoding="utf-8") as f:
        f.write(errorous_line)

    new_session = application.recover_session(
        session_id=session.session_id, filename=model_file
    )
    await new_session.get_service("event_manager").gather_tasks()
    new_element_factory = new_session.get_service("element_factory")

    assert not new_element_factory.lookup(diagram.id)
    assert "Could not recover model changes" in caplog.text


@pytest.mark.asyncio
@pytest.mark.parametrize("template", [True, False])
async def test_recover_from_session_files(
    application: Application, test_models, template
):
    session_id = "1234"
    class_id = "9876"

    create_recovery_file(
        session_id,
        {
            "path": str(test_models / "all-elements.gaphor"),
            "sha256": sha256sum(test_models / "all-elements.gaphor"),
            "template": template,
        },
        [("c", "UML", "Class", class_id, None)],
    )
    recover_sessions(application)
    for session in application.sessions:
        await session.get_service("event_manager").gather_tasks()

    session = next(s for s in application.sessions if s.session_id == session_id)
    element_factory = session.get_service("element_factory")

    assert element_factory.lookup(class_id)


def test_recover_with_invalid_filename(application: Application):
    session_id = "1234"
    class_id = "9876"

    create_recovery_file(
        session_id,
        {
            "path": "not-a-model.gaphor",
            "sha256": "1234",
            "template": False,
        },
        [("c", "UML", "Class", class_id, None)],
    )
    recover_sessions(application)

    assert not application.sessions


@pytest.mark.asyncio
async def test_recover_with_invalid_sha(application: Application, test_models):
    session_id = "1234"
    class_id = "9876"

    create_recovery_file(
        session_id,
        {
            "path": str(test_models / "all-elements.gaphor"),
            "sha256": "invalid-sha",
            "template": False,
        },
        [("c", "Class", class_id, None)],
    )
    recover_sessions(application)

    session = next(s for s in application.sessions if s.session_id == session_id)
    await session.get_service("event_manager").gather_tasks()
    element_factory = session.get_service("element_factory")

    assert not element_factory.lookup(class_id)


def test_recover_with_invalid_header(application: Application, test_models):
    session_id = "1234"
    class_id = "9876"

    create_recovery_file(
        session_id,
        "invalid header",
        [("c", "Class", class_id, None)],
    )
    recover_sessions(application)

    assert not application.sessions


def test_recover_with_unparseable_header(application: Application, test_models, caplog):
    session_id = "1234"
    class_id = "9876"

    create_recovery_file(
        session_id,
        {
            "path": str(test_models / "all-elements.gaphor"),
            "sha256": sha256sum(test_models / "all-elements.gaphor"),
            "template": True,
        },
        [("c", "Class", class_id, None)],
        raw_prefix="invalid",
    )
    recover_sessions(application)

    assert not application.sessions


def create_recovery_file(session_id, *lines, raw_prefix=""):
    with (sessions_dir() / f"{session_id}.recovery").open("w", encoding="utf-8") as f:
        f.write(raw_prefix)
        f.writelines(f"{repr(stmt)}\n" for stmt in lines)


@pytest.mark.asyncio
async def test_recovery_with_unlinked_item(application: Application, test_models):
    model_file = test_models / "simple-items.gaphor"
    session = application.new_session(template=model_file)
    event_manager = session.get_service("event_manager")
    await event_manager.gather_tasks()
    element_factory = session.get_service("element_factory")
    diagram = next(element_factory.select(Diagram))

    with Transaction(event_manager):
        line = diagram.create(Line)

    with Transaction(event_manager):
        line.handles()[0].pos = (100, 100)

    with Transaction(event_manager):
        line.unlink()

    application.shutdown_session(session)

    recover_sessions(application)

    assert application.sessions


@pytest.mark.asyncio
async def test_recovery_with_line_segments(application: Application, test_models):
    model_file = test_models / "simple-items.gaphor"
    session = application.new_session(template=model_file)
    event_manager = session.get_service("event_manager")
    await event_manager.gather_tasks()
    element_factory = session.get_service("element_factory")
    diagram = next(element_factory.select(Diagram))

    with Transaction(event_manager):
        line = diagram.create(Line)
        Segment(line, diagram).split_segment(0)

    with Transaction(event_manager):
        line.handles()[-1].pos = (100, 100)

    with Transaction(event_manager):
        Segment(line, diagram).merge_segment(0)

    application.shutdown_session(session)

    recover_sessions(application)

    assert application.sessions



import pytest

from gaphor import UML, SysML
from gaphor.diagram.presentation import DEFAULT_HEIGHT
from gaphor.diagram.propertypages import PropertyPages
from gaphor.diagram.tests.fixtures import find
from gaphor.SysML.propertypages import (
    CompartmentPage,
    ItemFlowPropertyPage,
    PropertyAggregationPropertyPage,
    RequirementItemPropertyPage,
    RequirementPropertyPage,
)
from gaphor.UML.propertypages import (
    ShowTypedElementPropertyPage,
    TypedElementPropertyPage,
)


@pytest.fixture
def association(element_factory):
    class_a = element_factory.create(UML.Class)
    class_b = element_factory.create(UML.Class)
    return UML.recipes.create_association(class_a, class_b)


@pytest.fixture
def connector(element_factory):
    prop_a = element_factory.create(UML.Property)
    prop_b = element_factory.create(UML.Property)
    return UML.recipes.create_connector(prop_a, prop_b)


def test_requirement_property_page_id(element_factory, event_manager):
    subject = element_factory.create(SysML.sysml.Requirement)
    property_page = RequirementPropertyPage(subject, event_manager)

    widget = property_page.construct()
    requirement_id = find(widget, "requirement-id")
    requirement_id.set_text("test")

    assert subject.externalId == "test"


def test_requirement_property_page_text(element_factory, event_manager):
    subject = element_factory.create(SysML.sysml.Requirement)
    property_page = RequirementPropertyPage(subject, event_manager)

    widget = property_page.construct()
    requirement_text = find(widget, "requirement-text")
    requirement_text.get_buffer().set_text("test")

    assert subject.text == "test"


def test_requirement_property_page_show_text(diagram, element_factory, event_manager):
    item = diagram.create(
        SysML.requirements.RequirementItem,
        subject=element_factory.create(SysML.sysml.Requirement),
    )
    property_page = RequirementItemPropertyPage(item, event_manager)

    widget = property_page.construct()
    show_requirement_text = find(widget, "show-requirement-text")

    assert item.show_text
    show_requirement_text.set_active(False)
    assert ~item.show_text


def test_requirement_height_reset(diagram, element_factory):
    item = diagram.create(
        SysML.requirements.RequirementItem,
        subject=element_factory.create(SysML.sysml.Requirement),
    )
    item.height = 500
    assert item.height == 500

    item.subject.text = "test"
    assert item.height == DEFAULT_HEIGHT


def test_show_property_type_property_page_show_type(
    diagram, element_factory, event_manager
):
    item = diagram.create(
        SysML.blocks.ProxyPortItem,
        subject=element_factory.create(SysML.sysml.ProxyPort),
    )
    property_page = ShowTypedElementPropertyPage(item, event_manager)

    widget = property_page.construct()
    show_parts = find(widget, "show-type")
    show_parts.set_active(True)

    assert item.show_type


def test_compartment_property_page_show_parts(diagram, element_factory, event_manager):
    item = diagram.create(
        SysML.blocks.BlockItem, subject=element_factory.create(SysML.sysml.Block)
    )
    property_page = CompartmentPage(item, event_manager)

    widget = property_page.construct()
    show_parts = find(widget, "show-parts")
    show_parts.set_active(True)

    assert item.show_parts


def test_compartment_property_page_show_references(
    diagram, element_factory, event_manager
):
    item = diagram.create(
        SysML.blocks.BlockItem, subject=element_factory.create(SysML.sysml.Block)
    )
    property_page = CompartmentPage(item, event_manager)

    widget = property_page.construct()
    show_references = find(widget, "show-references")
    show_references.set_active(True)

    assert item.show_references


def test_compartment_property_page_show_values(diagram, element_factory, event_manager):
    item = diagram.create(
        SysML.blocks.BlockItem, subject=element_factory.create(SysML.sysml.Block)
    )
    property_page = CompartmentPage(item, event_manager)

    widget = property_page.construct()
    show_references = find(widget, "show-values")
    show_references.set_active(True)

    assert item.show_values


def test_property_aggregation_page(element_factory, event_manager):
    subject = element_factory.create(UML.Property)
    property_page = PropertyAggregationPropertyPage(subject, event_manager)

    widget = property_page.construct()
    show_references = find(widget, "aggregation")
    show_references.set_selected(2)

    assert subject.aggregation == "composite"


def test_no_property_aggregation_page_for_ports(element_factory, event_manager):
    subject = element_factory.create(UML.Port)
    property_page = PropertyAggregationPropertyPage(subject, event_manager)

    widget = property_page.construct()

    assert not widget


def test_property_type(element_factory, event_manager):
    subject = element_factory.create(UML.Property)

    type = element_factory.create(UML.Class)
    type.name = "Bar"
    property_page = TypedElementPropertyPage(subject, event_manager)

    widget = property_page.construct()
    dropdown = find(widget, "element-type")
    bar_index = next(
        n for n, lv in enumerate(dropdown.get_model()) if lv.value == type.id
    )
    dropdown.set_selected(bar_index)

    assert dropdown.get_selected_item().label == "Bar"
    assert subject.type is type


def test_association_item_flow(association, event_manager):
    property_page = ItemFlowPropertyPage(association, event_manager)

    widget = property_page.construct()
    use = find(widget, "use-item-flow")
    use.set_active(True)

    assert association.abstraction
    assert association.abstraction[0].itemProperty
    assert association.abstraction[0].informationSource is association.memberEnd[0]
    assert association.abstraction[0].informationTarget is association.memberEnd[1]


def test_connector_item_flow(connector, event_manager):
    property_page = ItemFlowPropertyPage(connector, event_manager)

    widget = property_page.construct()
    use = find(widget, "use-item-flow")
    use.set_active(True)

    assert connector.informationFlow
    assert connector.informationFlow[0].itemProperty


def test_disable_item_flow(element_factory, event_manager):
    subject = element_factory.create(UML.Connector)
    property_page = ItemFlowPropertyPage(subject, event_manager)

    widget = property_page.construct()
    use = find(widget, "use-item-flow")
    use.set_active(True)
    use.set_active(False)

    assert not subject.informationFlow


def test_item_flow_name(connector, event_manager):
    property_page = ItemFlowPropertyPage(connector, event_manager)

    widget = property_page.construct()
    use = find(widget, "use-item-flow")
    use.set_active(True)
    entry = find(widget, "item-flow-name")
    entry.set_text("foo")

    assert entry.get_text() == "foo"
    assert connector.informationFlow[0].itemProperty.name == "foo"


def test_item_flow_type(connector, element_factory, event_manager):
    type = element_factory.create(UML.Class)
    type.name = "Bar"
    property_page = ItemFlowPropertyPage(connector, event_manager)

    widget = property_page.construct()
    use = find(widget, "use-item-flow")
    use.set_active(True)
    dropdown = find(widget, "item-flow-type")
    bar_index = next(
        n for n, lv in enumerate(dropdown.get_model()) if lv.value == type.id
    )
    dropdown.set_selected(bar_index)

    assert dropdown.get_selected_item().label == "Bar"
    assert connector.informationFlow[0].itemProperty.type is type


def test_item_flow_is_loaded(element_factory):
    subject = element_factory.create(UML.Connector)
    property_pages = PropertyPages.find(subject)

    assert any(issubclass(page, ItemFlowPropertyPage) for page in property_pages)


def test_item_flow_source_and_target(element_factory, event_manager):
    subject = element_factory.create(UML.Connector)
    head_end = element_factory.create(UML.ConnectorEnd)
    head_end.role = element_factory.create(UML.Property)
    tail_end = element_factory.create(UML.ConnectorEnd)
    tail_end.role = element_factory.create(UML.Property)
    subject.end = head_end
    subject.end = tail_end

    property_page = ItemFlowPropertyPage(subject, event_manager)

    widget = property_page.construct()
    use = find(widget, "use-item-flow")
    use.set_active(True)

    assert subject.informationFlow[0].informationSource is head_end.role
    assert subject.informationFlow[0].informationTarget is tail_end.role


def test_connector_item_flow_invert_direction(connector, event_manager):
    property_page = ItemFlowPropertyPage(connector, event_manager)

    widget = property_page.construct()
    use = find(widget, "use-item-flow")
    use.set_active(True)

    property_page.on_invert_direction_changed(None)

    information_flow = connector.informationFlow[0]
    assert information_flow.informationSource is connector.end[1].role
    assert information_flow.informationTarget is connector.end[0].role


def test_association_item_flow_invert_direction(association, event_manager):
    property_page = ItemFlowPropertyPage(association, event_manager)

    widget = property_page.construct()
    use = find(widget, "use-item-flow")
    use.set_active(True)

    property_page.on_invert_direction_changed(None)

    information_flow = association.abstraction[0]
    assert information_flow.informationSource is association.memberEnd[1]
    assert information_flow.informationTarget is association.memberEnd[0]



import pytest

from gaphor.diagram.general.modelinglanguage import GeneralModelingLanguage


def test_name():
    modeling_language = GeneralModelingLanguage()

    assert modeling_language.name == ""


def test_lookup_element():
    modeling_language = GeneralModelingLanguage()

    assert modeling_language.lookup_element("Box")
    assert modeling_language.lookup_element("Line")
    assert modeling_language.lookup_element("Ellipse")
    assert modeling_language.lookup_element("MetadataItem")


def test_toolbox():
    modeling_language = GeneralModelingLanguage()

    with pytest.raises(ValueError):
        modeling_language.toolbox_definition  # noqa: B018



from gaphor import UML
from gaphor.diagram.group import group, ungroup


def test_state_in_state_machine(element_factory):
    s = element_factory.create(UML.State)
    sm = element_factory.create(UML.StateMachine)

    group(sm, s)

    assert s.container in sm.region


def test_pseudostate_in_state_machine(element_factory):
    s = element_factory.create(UML.Pseudostate)
    sm = element_factory.create(UML.StateMachine)

    group(sm, s)

    assert s.container in sm.region


def test_state_in_region(element_factory):
    s = element_factory.create(UML.State)
    region = element_factory.create(UML.Region)

    group(region, s)

    assert s.container is region
    assert s in region.subvertex


def test_state_in_state(element_factory):
    s = element_factory.create(UML.State)
    parent = element_factory.create(UML.State)

    group(parent, s)

    assert s.container in parent.region


def test_ungroup_state_in_state_machine(element_factory):
    s = element_factory.create(UML.State)
    sm = element_factory.create(UML.StateMachine)

    group(sm, s)
    ungroup(sm, s)

    assert not s.container
    assert sm.region


def test_ungroup_pseudostate_in_state_machine(element_factory):
    s = element_factory.create(UML.Pseudostate)
    sm = element_factory.create(UML.StateMachine)

    group(sm, s)
    ungroup(sm, s)

    assert not s.container
    assert sm.region


def test_ungroup_state_in_region(element_factory):
    s = element_factory.create(UML.State)
    region = element_factory.create(UML.Region)

    group(region, s)
    ungroup(region, s)

    assert not s.container


def test_ungroup_state_in_state(element_factory):
    s = element_factory.create(UML.State)
    parent = element_factory.create(UML.State)

    group(parent, s)
    ungroup(parent, s)

    assert not s.container
    assert parent.region



import pytest

from gaphor.core.eventmanager import event_handler


class Event:
    pass


class OtherEvent:
    pass


@event_handler(Event)
class Subscriber:
    def __init__(self, exception=None):
        self.events = []
        self.exception = exception

    def __call__(self, event):
        self.events.append(event)
        if self.exception:
            raise self.exception()


def create_handler(event_type):
    events = []

    @event_handler(event_type)
    def handler(event):
        events.append(event)

    return handler, events


@pytest.fixture
def subscriber(event_manager):
    s = Subscriber()
    event_manager.subscribe(s)
    return s


def test_event_manager(event_manager, subscriber):
    event = Event()

    event_manager.handle(event)

    assert event in subscriber.events


def test_error_in_handler(event_manager):
    event = Event()
    error = Subscriber(ValueError)
    after = Subscriber()

    event_manager.subscribe(error)
    event_manager.subscribe(after)

    with pytest.raises(ExceptionGroup):
        event_manager.handle(event)

    assert event in error.events
    assert event in after.events


def test_error_in_handler_and_second_event(event_manager):
    @event_handler(Event)
    def error_handler(event):
        event_manager.handle(OtherEvent())
        raise ValueError()

    other_handler, other_events = create_handler(OtherEvent)

    event_manager.subscribe(error_handler)
    event_manager.subscribe(other_handler)

    event = Event()

    with pytest.raises(ExceptionGroup):
        event_manager.handle(event)

    assert not other_events


def test_priority_handler_error_in_handler_and_second_event(event_manager):
    @event_handler(Event)
    def error_handler(event):
        event_manager.handle(OtherEvent())
        raise ValueError()

    other_handler, other_events = create_handler(OtherEvent)

    event_manager.subscribe(error_handler)
    event_manager.priority_subscribe(other_handler)

    event = Event()

    with pytest.raises(ExceptionGroup):
        event_manager.handle(event)

    assert other_events


@pytest.mark.asyncio
async def test_async_event_handler(event_manager):
    events = []

    @event_handler(Event)
    async def handler(event):
        events.append(event)

    event_manager.subscribe(handler)

    event = Event()

    event_manager.handle(event)
    await event_manager.gather_tasks()

    assert event in events



import subprocess
import sys
from pathlib import Path
from urllib import error, request

import pytest

from gaphor.entrypoint import load_entry_points
from gaphor.plugins import (
    default_plugin_path,
    enable_plugins,
)


def have_internet():
    try:
        request.urlopen("http://gaphor.org", timeout=1)
    except error.URLError:
        return False
    else:
        return True


@pytest.mark.skipif(not have_internet(), reason="This test requires access to pypi")
def test_loading_plugins(tmp_path, monkeypatch):
    monkeypatch.setenv("GAPHOR_PLUGIN_PATH", str(tmp_path))

    subprocess.run(
        [
            sys.executable,
            "-m",
            "pip",
            "install",
            "--target",
            str(tmp_path),
            Path(__file__).parent.parent / "test-plugin",
        ],
        check=True,
    )

    with enable_plugins(default_plugin_path()):
        entry_points = load_entry_points("gaphor.services")

    assert "test_plugin" in entry_points



import sys

import pytest

from gaphor.ui.help.debuginfo import DebugInfo


class ApplicationStub:
    @property
    def active_session(self):
        return None


@pytest.fixture
def debug_info():
    svc = DebugInfo(ApplicationStub())
    yield svc
    svc.shutdown()


def exc_info():
    try:
        raise ValueError("Hello")
    except ValueError:
        return sys.exc_info()


def exc_group_info():
    def nested():
        raise ValueError("Internal Error")

    try:
        nested()
    except ValueError as e:
        exc = e
    try:
        raise ExceptionGroup("Group", [exc, exc])
    except ExceptionGroup:
        return sys.exc_info()


def test_exception_stored(debug_info):
    debug_info.excepthook(*exc_info())

    assert debug_info.exceptions


def test_format_exception(debug_info):
    debug_info.excepthook(*exc_info())
    text = debug_info.create_debug_info()

    assert 'raise ValueError("Hello")' in text


def test_format_exception_group(debug_info):
    debug_info.excepthook(*exc_group_info())
    text = debug_info.create_debug_info()

    assert 'ValueError("Internal Error")' in text



from gaphor.diagram.group import group
from gaphor.UML.uml import Interaction, Lifeline


def test_lifeline_group(element_factory):
    lifeline = element_factory.create(Lifeline)
    interaction = element_factory.create(Interaction)

    assert group(interaction, lifeline)



# ruff: noqa T201
#
# Update release info in appdata based  on Change log.

import subprocess
import sys
import textwrap
import time
from pathlib import Path
from xml.etree import ElementTree as etree

from markdown_it import MarkdownIt


def update_release(appdata_file, version, date, notes):
    tree = etree.parse(appdata_file)

    release = find_release(tree, version, date)
    update_release_details(release, version, notes)

    etree.indent(tree)
    tree.write(appdata_file, encoding="utf-8", xml_declaration=True)


def find_release(tree, version, date):
    if (release := tree.find(f"./releases/release[@version='{version}']")) is None:
        release = etree.Element("release", attrib={"version": version, "date": date})
        tree.find("./releases").insert(0, release)

    return release


def test_find_release():
    tree = etree.parse("data/org.gaphor.Gaphor.appdata.xml")

    release = find_release(tree, "2.17.0", "DATE")

    # Should have found the existing node
    assert release.attrib["date"] != "DATE"


def update_release_details(release, version, notes):
    for node in release.findall("*"):
        release.remove(node)

    release.append(notes)

    url = etree.SubElement(release, "url")
    url.text = f"https://github.com/gaphor/gaphor/releases/tag/{version}"


def parse_changelog(changelog_md: str):
    md = MarkdownIt("default", {"breaks": False, "html": True})
    html = md.render(changelog_md)
    changelog = etree.fromstring(f"<root>{html}</root>")

    version = "INVALID"
    news: dict[str, etree.Element] = {}

    for node in changelog:
        if node.tag == "h2" and node.text:
            version = node.text
            news[version] = etree.Element("description")
        else:
            news[version].append(node)

    return news


def test_parse_changelog():
    fragment = textwrap.dedent("""\
        2.1.0
        ------
        - Feature
          three
        - Feature four

        2.0.0
        ------
        - Feature one
        - Feature two
        """)

    changelog = parse_changelog(fragment)

    assert "2.0.0" in changelog
    assert "2.1.0" in changelog

    assert "Feature one" in etree.tostring(changelog["2.0.0"]).decode("utf-8")
    assert "Feature two" in etree.tostring(changelog["2.0.0"]).decode("utf-8")
    assert "Feature\nthree" in etree.tostring(changelog["2.1.0"]).decode("utf-8")
    assert "Feature four" in etree.tostring(changelog["2.1.0"]).decode("utf-8")


def run_tests():
    for name, maybe_test in list(globals().items()):
        if name.startswith("test_"):
            print(name, "...", end="")
            maybe_test()
            print(" okay")


if __name__ == "__main__":
    if len(sys.argv) > 1 and sys.argv[1] == "--test":
        run_tests()
    else:
        version = subprocess.run(
            ["poetry", "version", "-s"],
            encoding="utf-8",
            capture_output=True,
            text=True,
            check=True,
        ).stdout.rstrip()
        today = time.strftime("%Y-%m-%d")
        changelog = parse_changelog(Path("CHANGELOG.md").read_text())
        update_release(
            "data/org.gaphor.Gaphor.appdata.xml",
            version,
            today,
            changelog.get(version, ["Bug fixes."]),
        )



import pytest

from gaphor import UML
from gaphor.UML.states.pseudostates import PSEUDOSTATE_SHAPE, PseudostateItem


def test_initial_pseudostate(create):
    item = create(PseudostateItem, UML.Pseudostate)
    assert "initial" == item.subject.kind


def test_history_pseudostate(create):
    """Test creation of initial pseudostate."""
    item = create(PseudostateItem, UML.Pseudostate)
    # history setting is done in the DiagramToolbox factory:
    item.subject.kind = "shallowHistory"
    assert "shallowHistory" == item.subject.kind


@pytest.mark.parametrize("kind", UML.Pseudostate.kind.type)
def test_all_pseudostates_are_supported(kind):
    assert kind in PSEUDOSTATE_SHAPE



import pytest

from gaphor import UML
from gaphor.ui.treesearch import search, sorted_tree_walker
from gaphor.UML.treemodel import TreeModel


@pytest.fixture
def tree_model(event_manager, element_factory):
    return TreeModel(event_manager, element_factory)


@pytest.fixture
def create(element_factory, tree_model):
    def _create(name, parent=None):
        klass = element_factory.create(UML.Class)
        klass.name = name
        if parent:
            klass.nestingClass = parent
        tree_model.add_element(klass)
        return klass

    return _create


def test_search(tree_model, create):
    create("aaa")
    bbb = create("bbb")

    found = search("b", sorted_tree_walker(tree_model))

    assert found.element is bbb


def test_search_no_hit(tree_model, create):
    create("aaa")
    create("bbb")

    found = search("z", sorted_tree_walker(tree_model))

    assert found is None


def test_search_with_child_models(tree_model, create):
    aaa = create("aaa")
    bbb = create("bbb", parent=aaa)

    found = search("b", sorted_tree_walker(tree_model))

    assert found.element is bbb


def test_search_with_start_tree_item(tree_model, create):
    create("aab")
    abb = create("abb")
    bbb = create("bbb")

    found = search(
        "b",
        sorted_tree_walker(
            tree_model, start_tree_item=tree_model.tree_item_for_element(abb)
        ),
    )

    assert found.element is bbb


def test_search_from_current_with_start_tree_item(tree_model, create):
    create("aab")
    abb = create("abb")
    create("bbb")

    found = search(
        "b",
        sorted_tree_walker(
            tree_model,
            start_tree_item=tree_model.tree_item_for_element(abb),
            from_current=True,
        ),
    )

    assert found.element is abb



# ruff: noqa: F401,F811
import pytest

from gaphor.UML import Comment
from gaphor.UML.general import CommentItem
from gaphor.UML.general.commenteditor import comment_item_editor


@pytest.mark.asyncio
async def test_comment_item_editor(diagram, element_factory, view, event_manager):
    item = diagram.create(CommentItem, subject=element_factory.create(Comment))
    view.selection.hovered_item = item
    await view.update()
    result = comment_item_editor(item, view, event_manager)

    assert result is True
